"""
Module which defines the code for the "Generate speech - multi-step
generation" tab.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from functools import partial

import gradio as gr

from ultimate_rvc.core.common import SPEECH_DIR
from ultimate_rvc.core.generate.common import convert
from ultimate_rvc.core.generate.speech import (
    get_mixed_speech_track_name,
    mix_speech,
    run_edge_tts,
)
from ultimate_rvc.core.manage.audio import (
    get_saved_output_audio,
    get_saved_speech_audio,
)
from ultimate_rvc.typing_extra import EmbedderModel, RVCContentType
from ultimate_rvc.web.common import (
    exception_harness,
    render_transfer_component,
    setup_transfer_event,
    toggle_visibility,
    toggle_visible_component,
    update_dropdowns,
    update_output_name,
    update_value,
)
from ultimate_rvc.web.typing_extra import ConcurrencyId, SpeechTransferOption

if TYPE_CHECKING:
    from ultimate_rvc.web.config.main import TotalConfig


def render(total_config: TotalConfig) -> None:
    """
    Render "Generate speech - multi-step generation" tab.

    Parameters
    ----------
    total_config : TotalConfig
        Model containing all component configuration settings for the
        Ultimate RVC web UI.

    """
    tab_config = total_config.speech.multi_step
    for c in tab_config.input_audio.all:
        c.instantiate()
    with gr.Tab("Multi-step"):
        _render_step_1(total_config)
        _render_step_2(total_config)
        _render_step_3(total_config)


def _render_step_1(total_config: TotalConfig) -> None:
    tab_config = total_config.speech.multi_step
    with gr.Accordion("Step 1: Text-to-speech conversion", open=True):
        with gr.Row():
            with gr.Column():
                tab_config.source_type.instantiate()
            with gr.Column():
                tab_config.source.instantiate()
                local_file = gr.File(
                    label="Source",
                    file_types=[".txt"],
                    file_count="single",
                    type="filepath",
                    visible=False,
                )
            tab_config.source_type.instance.input(
                partial(toggle_visible_component, 2),
                inputs=tab_config.source_type.instance,
                outputs=[tab_config.source.instance, local_file],
                show_progress="hidden",
            )
            local_file.change(
                update_value,
                inputs=local_file,
                outputs=tab_config.source.instance,
                show_progress="hidden",
            )
        tab_config.edge_tts_voice.instance.render()
        with gr.Accordion("Options", open=False):
            with gr.Row():
                tab_config.tts_pitch_shift.instantiate()
                tab_config.tts_speed_change.instantiate()
                tab_config.tts_volume_change.instantiate()
            speech_transfer = _render_speech_transfer(
                [SpeechTransferOption.STEP_2_SPEECH],
                "Speech",
            )
        with gr.Row():
            tts_reset_btn = gr.Button("Reset settings")
            tts_btn = gr.Button("Convert text", variant="primary")
        tts_transfer_btn = gr.Button("Transfer speech")

        speech_track_output = gr.Audio(
            label="Generated speech",
            type="filepath",
            interactive=False,
            waveform_options=gr.WaveformOptions(show_recording_waveform=False),
        )
        tts_reset_btn.click(
            lambda: [
                tab_config.tts_pitch_shift.value,
                tab_config.tts_speed_change.value,
                tab_config.tts_volume_change.value,
                gr.Dropdown(value=[SpeechTransferOption.STEP_2_SPEECH]),
            ],
            outputs=[
                tab_config.tts_pitch_shift.instance,
                tab_config.tts_speed_change.instance,
                tab_config.tts_volume_change.instance,
                speech_transfer,
            ],
            show_progress="hidden",
        )
        tts_btn.click(
            exception_harness(run_edge_tts, info_msg="Text succesfully converted!"),
            inputs=[
                tab_config.source.instance,
                tab_config.edge_tts_voice.instance,
                tab_config.tts_pitch_shift.instance,
                tab_config.tts_speed_change.instance,
                tab_config.tts_volume_change.instance,
            ],
            outputs=speech_track_output,
        ).then(
            partial(update_dropdowns, get_saved_speech_audio, 1, [], [0]),
            outputs=total_config.management.audio.speech.instance,
            show_progress="hidden",
        )
        setup_transfer_event(
            tts_transfer_btn,
            speech_transfer,
            speech_track_output,
            tab_config.input_audio.all,
        )


def _render_step_2(total_config: TotalConfig) -> None:
    tab_config = total_config.speech.multi_step

    with gr.Accordion("Step 2: speech conversion", open=False):
        tab_config.input_audio.speech.instance.render()
        tab_config.voice_model.instance.render()
        with gr.Accordion("Options", open=False):
            with gr.Row():
                tab_config.n_octaves.instantiate()
                tab_config.n_semitones.instantiate()
            with gr.Accordion("Voice synthesis settings", open=False):
                with gr.Row():
                    tab_config.f0_methods.instantiate()
                    tab_config.index_rate.instantiate()
                with gr.Row():
                    tab_config.rms_mix_rate.instantiate()
                    tab_config.protect_rate.instantiate()
                    tab_config.hop_length.instantiate()
            with gr.Accordion("Speech enrichment settings", open=False), gr.Row():
                with gr.Column():
                    tab_config.split_voice.instantiate()
                with gr.Column():
                    tab_config.autotune_voice.instantiate()
                    tab_config.autotune_strength.instantiate()
                with gr.Column():
                    tab_config.clean_voice.instantiate()
                    tab_config.clean_strength.instantiate()
            tab_config.autotune_voice.instance.change(
                partial(toggle_visibility, targets={True}),
                inputs=tab_config.autotune_voice.instance,
                outputs=tab_config.autotune_strength.instance,
                show_progress="hidden",
            )
            tab_config.clean_voice.instance.change(
                partial(toggle_visibility, targets={True}),
                inputs=tab_config.clean_voice.instance,
                outputs=tab_config.clean_strength.instance,
                show_progress="hidden",
            )
            with gr.Accordion("Speaker embedding settings", open=False), gr.Row():
                with gr.Column():
                    tab_config.embedder_model.instantiate()
                    tab_config.custom_embedder_model.instance.render()
                tab_config.sid.instantiate()
            tab_config.embedder_model.instance.change(
                partial(toggle_visibility, targets={EmbedderModel.CUSTOM}),
                inputs=tab_config.embedder_model.instance,
                outputs=tab_config.custom_embedder_model.instance,
                show_progress="hidden",
            )
            converted_speech_transfer = _render_speech_transfer(
                [SpeechTransferOption.STEP_3_SPEECH],
                "Converted speech",
            )
        with gr.Row():
            convert_speech_reset_btn = gr.Button("Reset settings")
            convert_speech_btn = gr.Button("Convert speech", variant="primary")
        converted_speech_transfer_btn = gr.Button("Transfer converted speech")

        converted_speech_track_output = gr.Audio(
            label="Converted speech",
            type="filepath",
            interactive=False,
            waveform_options=gr.WaveformOptions(show_recording_waveform=False),
        )
        convert_speech_reset_btn.click(
            lambda: [
                tab_config.n_octaves.value,
                tab_config.n_semitones.value,
                tab_config.f0_methods.value,
                tab_config.index_rate.value,
                tab_config.rms_mix_rate.value,
                tab_config.protect_rate.value,
                tab_config.hop_length.value,
                tab_config.split_voice.value,
                tab_config.autotune_voice.value,
                tab_config.autotune_strength.value,
                tab_config.clean_voice.value,
                tab_config.clean_strength.value,
                tab_config.embedder_model.value,
                tab_config.sid.value,
                gr.Dropdown(value=[SpeechTransferOption.STEP_3_SPEECH]),
            ],
            outputs=[
                tab_config.n_octaves.instance,
                tab_config.n_semitones.instance,
                tab_config.f0_methods.instance,
                tab_config.index_rate.instance,
                tab_config.rms_mix_rate.instance,
                tab_config.protect_rate.instance,
                tab_config.hop_length.instance,
                tab_config.split_voice.instance,
                tab_config.autotune_voice.instance,
                tab_config.autotune_strength.instance,
                tab_config.clean_voice.instance,
                tab_config.clean_strength.instance,
                tab_config.embedder_model.instance,
                tab_config.sid.instance,
                converted_speech_transfer,
            ],
            show_progress="hidden",
        )

        convert_speech_btn.click(
            partial(
                exception_harness(convert, info_msg="Speech succesfully converted!"),
                content_type=RVCContentType.SPEECH,
                make_directory=True,
            ),
            inputs=[
                tab_config.input_audio.speech.instance,
                gr.State(SPEECH_DIR),
                tab_config.voice_model.instance,
                tab_config.n_octaves.instance,
                tab_config.n_semitones.instance,
                tab_config.f0_methods.instance,
                tab_config.index_rate.instance,
                tab_config.rms_mix_rate.instance,
                tab_config.protect_rate.instance,
                tab_config.hop_length.instance,
                tab_config.split_voice.instance,
                tab_config.autotune_voice.instance,
                tab_config.autotune_strength.instance,
                tab_config.clean_voice.instance,
                tab_config.clean_strength.instance,
                tab_config.embedder_model.instance,
                tab_config.custom_embedder_model.instance,
                tab_config.sid.instance,
            ],
            outputs=converted_speech_track_output,
            concurrency_id=ConcurrencyId.GPU,
            concurrency_limit=1,
        ).then(
            partial(update_dropdowns, get_saved_speech_audio, 1, [], [0]),
            outputs=total_config.management.audio.speech.instance,
            show_progress="hidden",
        )
        setup_transfer_event(
            converted_speech_transfer_btn,
            converted_speech_transfer,
            converted_speech_track_output,
            tab_config.input_audio.all,
        )


def _render_step_3(total_config: TotalConfig) -> None:
    tab_config = total_config.speech.multi_step
    with gr.Accordion("Step 3: speech mixing", open=False):
        tab_config.input_audio.converted_speech.instance.render()
        with gr.Accordion("Options", open=False):
            with gr.Row():
                tab_config.output_gain.instantiate()
                tab_config.output_sr.instantiate()
            with gr.Row():
                tab_config.output_name.instantiate(
                    value=partial(
                        update_output_name,
                        get_mixed_speech_track_name,
                        False,  # noqa: FBT003
                    ),
                    inputs=[
                        gr.State(None),
                        gr.State(None),
                        tab_config.input_audio.converted_speech.instance,
                    ],
                )
                tab_config.output_format.instantiate()
            mixed_speech_transfer = _render_speech_transfer([], "Mixed speech")
        with gr.Row():
            mix_speech_btn = gr.Button("Mix speech", variant="primary")
            mix_speech_transfer_btn = gr.Button("Transfer mixed speech")
        mix_speech_reset_btn = gr.Button("Reset settings")
        mixed_speech_track_output = gr.Audio(
            label="Mixed speech",
            type="filepath",
            interactive=False,
            waveform_options=gr.WaveformOptions(show_recording_waveform=False),
        )

        mix_speech_reset_btn.click(
            lambda: [
                tab_config.output_gain.value,
                tab_config.output_sr.value,
                tab_config.output_format.value,
                gr.Dropdown(value=[]),
            ],
            outputs=[
                tab_config.output_gain.instance,
                tab_config.output_sr.instance,
                tab_config.output_format.instance,
                mixed_speech_transfer,
            ],
            show_progress="hidden",
        )
        mix_speech_btn.click(
            exception_harness(mix_speech, info_msg="Speech successfully mixed!"),
            inputs=[
                tab_config.input_audio.converted_speech.instance,
                tab_config.output_gain.instance,
                tab_config.output_sr.instance,
                tab_config.output_format.instance,
                tab_config.output_name.instance,
            ],
            outputs=mixed_speech_track_output,
        ).then(
            partial(update_dropdowns, get_saved_output_audio, 1, [], [0]),
            outputs=total_config.management.audio.output.instance,
            show_progress="hidden",
        )
        setup_transfer_event(
            mix_speech_transfer_btn,
            mixed_speech_transfer,
            mixed_speech_track_output,
            tab_config.input_audio.all,
        )


def _render_speech_transfer(
    value: list[SpeechTransferOption],
    label_prefix: str,
) -> gr.Dropdown:
    return render_transfer_component(value, label_prefix, SpeechTransferOption)



"""
Module which defines the command-line interface for the Ultimate RVC
project.
"""

from __future__ import annotations

import typer

from ultimate_rvc.cli.generate.main import app as generate_app
from ultimate_rvc.cli.train.main import app as train_app

app = typer.Typer(
    name="urvc-cli",
    no_args_is_help=True,
    help="CLI for the Ultimate RVC project",
    rich_markup_mode="markdown",
)

app.add_typer(generate_app)
app.add_typer(train_app)


if __name__ == "__main__":
    app()



"""Module which defines functions to manage audio files."""

from __future__ import annotations

from typing import TYPE_CHECKING

import shutil
from pathlib import Path

from ultimate_rvc.common import AUDIO_DIR
from ultimate_rvc.core.common import (
    INTERMEDIATE_AUDIO_BASE_DIR,
    OUTPUT_AUDIO_DIR,
    SPEECH_DIR,
    TRAINING_AUDIO_DIR,
)
from ultimate_rvc.core.exceptions import (
    Entity,
    InvalidLocationError,
    Location,
    NotFoundError,
    NotProvidedError,
    UIMessage,
)
from ultimate_rvc.core.manage.common import (
    delete_directory,
    get_items,
    get_named_items,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ultimate_rvc.typing_extra import StrPath


def get_saved_output_audio() -> list[tuple[str, str]]:
    """
    Get the name and path of all output audio files.

    Returns
    -------
    list[tuple[str, str]]
        A list of tuples containing the name and path of each output
        audio file.

    """
    return get_named_items(OUTPUT_AUDIO_DIR)


def get_saved_speech_audio() -> list[tuple[str, str]]:
    """
    Get the name and path of all speech audio files.

    Returns
    -------
    list[tuple[str, Path]]
        A list of tuples containing the name and path of each
        speech audio file.

    """
    return get_named_items(SPEECH_DIR, exclude=".json")


def get_named_audio_datasets() -> list[tuple[str, str]]:
    """
    Get the name and path of all saved audio datasets.

    Returns
    -------
    list[tuple[str, str]]
        A list of tuples containing the name and path of each saved
        audio dataset.

    """
    return get_named_items(TRAINING_AUDIO_DIR)


def get_audio_datasets() -> list[str]:
    """
    Get the paths of all saved audio datasets.

    Returns
    -------
    list[str]
        A list of the paths of all saved audio datasets.

    """
    return get_items(TRAINING_AUDIO_DIR, only_stem=False)


def delete_audio(
    directory: StrPath,
    items: Sequence[StrPath],
    plural_entity: Entity = Entity.FILES,
    singular_entity: Entity = Entity.FILE,
    location: Location = Location.AUDIO_ROOT,
    ui_msg: UIMessage = UIMessage.NO_UPLOADED_FILES,
) -> None:
    """
    Delete the provided audio items.

    The provided audio items must be located in the root of the provided
    directory.

    Parameters
    ----------
    directory : StrPath
        The path to the directory containing the audio items to delete
    items : Sequence[StrPath]
        Paths to the audio items to delete.
    plural_entity : Entity, optional
        The plural form of the entity being deleted.
    singular_entity : Entity, optional
        The singular form of the entity being deleted.
    location : Location, optional
        The location where the files should be located.
    ui_msg : UIMessage, optional
        The message to display if no items are provided.

    Raises
    ------
    NotProvidedError
        If no paths are provided.
    NotFoundError
        If a provided path does not point to an existing audio item.
    InvalidLocationError
        If a provided path does not point to a location in the root of
        the provided directory.

    """
    if not items:
        raise NotProvidedError(entity=plural_entity, ui_msg=ui_msg)
    item_paths: list[Path] = []
    for item in items:
        item_path = Path(item)
        if not item_path.exists():
            raise NotFoundError(entity=singular_entity, location=item_path)
        if item_path.parent != Path(directory):
            raise InvalidLocationError(singular_entity, location, item_path)
        item_paths.append(item_path)
    for item_path in item_paths:
        if item_path.is_dir():
            shutil.rmtree(item_path)
        else:
            item_path.unlink()
            json_path = item_path.with_suffix(".json")
            json_path.unlink(missing_ok=True)


def delete_intermediate_audio(directories: Sequence[StrPath]) -> None:
    """
    Delete provided directories containing intermediate audio files.

    The provided directories must be located in the root of the
    intermediate audio base directory.

    Parameters
    ----------
    directories : Sequence[StrPath]
        Paths to directories containing intermediate audio files to
        delete.

    """
    delete_audio(
        directory=INTERMEDIATE_AUDIO_BASE_DIR,
        items=directories,
        plural_entity=Entity.DIRECTORIES,
        singular_entity=Entity.DIRECTORY,
        location=Location.INTERMEDIATE_AUDIO_ROOT,
        ui_msg=UIMessage.NO_SONG_DIRS,
    )


def delete_speech_audio(files: Sequence[StrPath]) -> None:
    """
    Delete provided speech audio files.

    The provided files must be located in the root of the speech audio
    directory.

    Parameters
    ----------
    files : Sequence[StrPath]
        Paths to the speech audio files to delete.

    """
    delete_audio(
        directory=SPEECH_DIR,
        items=files,
        plural_entity=Entity.FILES,
        singular_entity=Entity.FILE,
        location=Location.SPEECH_AUDIO_ROOT,
        ui_msg=UIMessage.NO_SPEECH_AUDIO_FILES,
    )


def delete_output_audio(files: Sequence[StrPath]) -> None:
    """
    Delete provided output audio files.

    The provided files must be located in the root of the output audio
    directory.

    Parameters
    ----------
    files : Sequence[StrPath]
        Paths to the output audio files to delete.

    """
    delete_audio(
        directory=OUTPUT_AUDIO_DIR,
        items=files,
        plural_entity=Entity.FILES,
        singular_entity=Entity.FILE,
        location=Location.OUTPUT_AUDIO_ROOT,
        ui_msg=UIMessage.NO_OUTPUT_AUDIO_FILES,
    )


def delete_dataset_audio(datasets: Sequence[StrPath]) -> None:
    """
    Delete provided datasets containing audio files.

    The provided datasets must be located in the root of the training
    audio directory.

    Parameters
    ----------
    datasets : Sequence[StrPath]
        Paths to the datasets to delete.

    """
    delete_audio(
        directory=TRAINING_AUDIO_DIR,
        items=datasets,
        plural_entity=Entity.DIRECTORIES,
        singular_entity=Entity.DIRECTORY,
        location=Location.TRAINING_AUDIO_ROOT,
        ui_msg=UIMessage.NO_DATASETS,
    )


def delete_all_intermediate_audio() -> None:
    """Delete all intermediate audio files."""
    delete_directory(INTERMEDIATE_AUDIO_BASE_DIR)


def delete_all_speech_audio() -> None:
    """Delete all speech audio files."""
    delete_directory(SPEECH_DIR)


def delete_all_output_audio() -> None:
    """Delete all output audio files."""
    delete_directory(OUTPUT_AUDIO_DIR)


def delete_all_dataset_audio() -> None:
    """Delete all dataset audio files."""
    delete_directory(TRAINING_AUDIO_DIR)


def delete_all_audio() -> None:
    """Delete all audio files."""
    delete_directory(AUDIO_DIR)



"""Module which defines the code for the "Audio" tab."""

from __future__ import annotations

from typing import TYPE_CHECKING

from functools import partial

import gradio as gr

from ultimate_rvc.core.generate.song_cover import get_named_song_dirs
from ultimate_rvc.core.manage.audio import (
    delete_all_audio,
    delete_all_dataset_audio,
    delete_all_intermediate_audio,
    delete_all_output_audio,
    delete_all_speech_audio,
    delete_dataset_audio,
    delete_intermediate_audio,
    delete_output_audio,
    delete_speech_audio,
    get_audio_datasets,
    get_named_audio_datasets,
    get_saved_output_audio,
    get_saved_speech_audio,
)
from ultimate_rvc.web.common import setup_delete_event, update_dropdowns
from ultimate_rvc.web.config.event import ManageAudioEventState

if TYPE_CHECKING:
    from ultimate_rvc.web.config.main import AudioManagementConfig, TotalConfig


def render(total_config: TotalConfig) -> None:
    """
    Render "Audio" tab.

    Parameters
    ----------
    total_config: TotalConfig
        Model containing all component configuration settings for the
        Ultimate RVC web UI.

    """
    tab_config = total_config.management.audio
    tab_config.dummy_checkbox.instantiate()
    event_state = ManageAudioEventState()

    with gr.Tab("Delete"):
        _render_intermediate_accordion(tab_config, event_state)
        _render_speech_accordion(tab_config, event_state)
        _render_output_accordion(tab_config, event_state)
        _render_dataset_accordion(tab_config, event_state)
        _render_all_accordion(tab_config, event_state)

        _, _, all_audio_update = [
            click_event.success(
                partial(
                    update_dropdowns,
                    get_named_song_dirs,
                    3 + len(total_config.song.multi_step.song_dirs.all),
                    [],
                    [0],
                ),
                outputs=[
                    tab_config.intermediate.instance,
                    total_config.song.one_click.cached_song.instance,
                    total_config.song.multi_step.cached_song.instance,
                    *total_config.song.multi_step.song_dirs.all,
                ],
                show_progress="hidden",
            )
            for click_event in [
                event_state.delete_intermediate_click.instance,
                event_state.delete_all_intermediate_click.instance,
                event_state.delete_all_click.instance,
            ]
        ]

        _, _, all_audio_update = [
            click_event.success(
                partial(update_dropdowns, get_saved_speech_audio, 1, [], [0]),
                outputs=tab_config.speech.instance,
                show_progress="hidden",
            )
            for click_event in [
                event_state.delete_speech_click.instance,
                event_state.delete_all_speech_click.instance,
                all_audio_update,
            ]
        ]

        _, _, all_audio_update = [
            click_event.success(
                partial(update_dropdowns, get_saved_output_audio, 1, [], [0]),
                outputs=tab_config.output.instance,
                show_progress="hidden",
            )
            for click_event in [
                event_state.delete_output_click.instance,
                event_state.delete_all_output_click.instance,
                all_audio_update,
            ]
        ]

        for click_event in [
            event_state.delete_dataset_click.instance,
            event_state.delete_all_dataset_click.instance,
            all_audio_update,
        ]:
            click_event.success(
                partial(update_dropdowns, get_named_audio_datasets, 1, [], [0]),
                outputs=tab_config.dataset.instance,
                show_progress="hidden",
            ).then(
                partial(update_dropdowns, get_audio_datasets, 1, [], [0]),
                outputs=total_config.training.multi_step.dataset.instance,
                show_progress="hidden",
            )


def _render_intermediate_accordion(
    tab_config: AudioManagementConfig,
    event_state: ManageAudioEventState,
) -> None:
    with gr.Accordion("Intermediate audio", open=False), gr.Row():
        with gr.Column():
            tab_config.intermediate.instance.render()
            intermediate_audio_btn = gr.Button("Delete selected", variant="secondary")
            all_intermediate_audio_btn = gr.Button("Delete all", variant="primary")
        with gr.Column():
            intermediate_audio_msg = gr.Textbox(
                label="Output message",
                interactive=False,
            )
        event_state.delete_intermediate_click.instance = setup_delete_event(
            intermediate_audio_btn,
            delete_intermediate_audio,
            [tab_config.dummy_checkbox.instance, tab_config.intermediate.instance],
            intermediate_audio_msg,
            "Are you sure you want to delete the selected song directories?",
            "[-] Successfully deleted the selected song directories!",
        )
        event_state.delete_all_intermediate_click.instance = setup_delete_event(
            all_intermediate_audio_btn,
            delete_all_intermediate_audio,
            [tab_config.dummy_checkbox.instance],
            intermediate_audio_msg,
            "Are you sure you want to delete all intermediate audio files?",
            "[-] Successfully deleted all intermediate audio files!",
        )


def _render_speech_accordion(
    tab_config: AudioManagementConfig,
    event_state: ManageAudioEventState,
) -> None:
    with gr.Accordion("Speech audio", open=False), gr.Row():
        with gr.Column():
            tab_config.speech.instance.render()
            speech_audio_btn = gr.Button("Delete selected", variant="secondary")
            all_speech_audio_btn = gr.Button("Delete all", variant="primary")
        with gr.Column():
            speech_audio_msg = gr.Textbox(label="Output message", interactive=False)

        event_state.delete_speech_click.instance = setup_delete_event(
            speech_audio_btn,
            delete_speech_audio,
            [tab_config.dummy_checkbox.instance, tab_config.speech.instance],
            speech_audio_msg,
            "Are you sure you want to delete the selected speech audio files?",
            "[-] Successfully deleted the selected speech audio files!",
        )

        event_state.delete_all_speech_click.instance = setup_delete_event(
            all_speech_audio_btn,
            delete_all_speech_audio,
            [tab_config.dummy_checkbox.instance],
            speech_audio_msg,
            "Are you sure you want to delete all speech audio files?",
            "[-] Successfully deleted all speech audio files!",
        )


def _render_output_accordion(
    tab_config: AudioManagementConfig,
    event_state: ManageAudioEventState,
) -> None:
    with gr.Accordion("Output audio", open=False), gr.Row():
        with gr.Column():
            tab_config.output.instance.render()
            output_audio_btn = gr.Button("Delete selected", variant="secondary")
            all_output_audio_btn = gr.Button("Delete all", variant="primary")
        with gr.Column():
            output_audio_msg = gr.Textbox(label="Output message", interactive=False)
        event_state.delete_output_click.instance = setup_delete_event(
            output_audio_btn,
            delete_output_audio,
            [tab_config.dummy_checkbox.instance, tab_config.output.instance],
            output_audio_msg,
            "Are you sure you want to delete the selected output audio files?",
            "[-] Successfully deleted the selected output audio files!",
        )
        event_state.delete_all_output_click.instance = setup_delete_event(
            all_output_audio_btn,
            delete_all_output_audio,
            [tab_config.dummy_checkbox.instance],
            output_audio_msg,
            "Are you sure you want to delete all output audio files?",
            "[-] Successfully deleted all output audio files!",
        )


def _render_dataset_accordion(
    tab_config: AudioManagementConfig,
    event_state: ManageAudioEventState,
) -> None:
    with gr.Accordion("Dataset audio", open=False), gr.Row():
        with gr.Column():
            tab_config.dataset.instance.render()
            dataset_audio_btn = gr.Button("Delete selected", variant="secondary")
            all_dataset_audio_btn = gr.Button("Delete all", variant="primary")
        with gr.Column():
            dataset_audio_msg = gr.Textbox(label="Output message", interactive=False)

        event_state.delete_dataset_click.instance = setup_delete_event(
            dataset_audio_btn,
            delete_dataset_audio,
            [tab_config.dummy_checkbox.instance, tab_config.dataset.instance],
            dataset_audio_msg,
            "Are you sure you want to delete the selected dataset audio files?",
            "[-] Successfully deleted the selected dataset audio files!",
        )
        event_state.delete_all_dataset_click.instance = setup_delete_event(
            all_dataset_audio_btn,
            delete_all_dataset_audio,
            [tab_config.dummy_checkbox.instance],
            dataset_audio_msg,
            "Are you sure you want to delete all dataset audio files?",
            "[-] Successfully deleted all dataset audio files!",
        )


def _render_all_accordion(
    tab_config: AudioManagementConfig,
    event_state: ManageAudioEventState,
) -> None:
    with gr.Accordion("All audio", open=True), gr.Row(equal_height=True):
        all_audio_btn = gr.Button("Delete", variant="primary")
        all_audio_msg = gr.Textbox(label="Output message", interactive=False)

    event_state.delete_all_click.instance = setup_delete_event(
        all_audio_btn,
        delete_all_audio,
        [tab_config.dummy_checkbox.instance],
        all_audio_msg,
        "Are you sure you want to delete all audio files?",
        "[-] Successfully deleted all audio files!",
    )



"""Common utilities for the CLI."""

from ultimate_rvc.typing_extra import (
    AudioExt,
    AudioSplitMethod,
    DeviceType,
    EmbedderModel,
    F0Method,
    IndexAlgorithm,
    PretrainedType,
    TrainingSampleRate,
    Vocoder,
)


def format_duration(seconds: float) -> str:
    """
    Format a duration in seconds to a human-readable string.

    Parameters
    ----------
    seconds : float
        The duration in seconds.

    Returns
    -------
    str
        The formatted duration

    """
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    if hours > 0:
        return f"{int(hours)} hours, {int(minutes)} minutes, and {seconds:.2f} seconds"
    if minutes > 0:
        return f"{int(minutes)} minutes and {seconds:.2f} seconds"
    return f"{seconds:.2f} seconds"


def complete_name(incomplete: str, enumeration: list[str]) -> list[str]:
    """
    Return a list of names that start with the incomplete string.

    Parameters
    ----------
    incomplete : str
        The incomplete string to complete.
    enumeration : list[str]
        The list of names to complete from.

    Returns
    -------
    list[str]
        The list of names that start with the incomplete string.

    """
    return [name for name in list(enumeration) if name.startswith(incomplete)]


def complete_audio_ext(incomplete: str) -> list[str]:
    """
    Return a list of audio extensions that start with the incomplete
    string.

    Parameters
    ----------
    incomplete : str
        The incomplete string to complete.

    Returns
    -------
    list[str]
        The list of audio extensions that start with the incomplete
        string.

    """
    return complete_name(incomplete, list(AudioExt))


def complete_f0_method(incomplete: str) -> list[str]:
    """
    Return a list of F0 methods that start with the incomplete string.

    Parameters
    ----------
    incomplete : str
        The incomplete string to complete.

    Returns
    -------
    list[str]
        The list of F0 methods that start with the incomplete string.

    """
    return complete_name(incomplete, list(F0Method))


def complete_embedder_model(incomplete: str) -> list[str]:
    """
    Return a list of embedder models that start with the incomplete
    string.

    Parameters
    ----------
    incomplete : str
        The incomplete string to complete.

    Returns
    -------
    list[str]
        The list of embedder models that start with the incomplete
        string.

    """
    return complete_name(incomplete, list(EmbedderModel))


def complete_audio_split_method(incomplete: str) -> list[str]:
    """
    Return a list of audio split methods that start with the incomplete
    string.

    Parameters
    ----------
    incomplete : str
        The incomplete string to complete.

    Returns
    -------
    list[str]
        The list of audio split methods that start with the incomplete
        string.

    """
    return complete_name(incomplete, list(AudioSplitMethod))


def complete_training_sample_rate(incomplete: str) -> list[str]:
    """
    Return a list of training sample rates that start with the
    incomplete string.

    Parameters
    ----------
    incomplete : str
        The incomplete string to complete.

    Returns
    -------
    list[str]
        The list of training sample rates that start with the incomplete
        string.

    """
    return complete_name(incomplete, list(TrainingSampleRate))


def complete_vocoder(incomplete: str) -> list[str]:
    """
    Return a list of vocoders that start with the incomplete string.

    Parameters
    ----------
    incomplete : str
        The incomplete string to complete.

    Returns
    -------
    list[str]
        The list of vocoders that start with the incomplete string.

    """
    return complete_name(incomplete, list(Vocoder))


def complete_index_algorithm(incomplete: str) -> list[str]:
    """
    Return a list of index algorithms that start with the incomplete
    string.

    Parameters
    ----------
    incomplete : str
        The incomplete string to complete.

    Returns
    -------
    list[str]
        The list of index algorithms that start with the incomplete
        string.

    """
    return complete_name(incomplete, list(IndexAlgorithm))


def complete_device_type(incomplete: str) -> list[str]:
    """
    Return a list of device types that start with the incomplete
    string.

    Parameters
    ----------
    incomplete : str
        The incomplete string to complete.

    Returns
    -------
    list[str]
        The list of device types that start with the incomplete
        string.

    """
    return complete_name(incomplete, list(DeviceType))


def complete_pretrained_type(incomplete: str) -> list[str]:
    """
    Return a list of pretrained model types that start with the
    incomplete string.

    Parameters
    ----------
    incomplete : str
        The incomplete string to complete.

    Returns
    -------
    list[str]
        The list of pretrained model types that start with the
        incomplete string.

    """
    return complete_name(incomplete, list(PretrainedType))



from typing import Self

from pathlib import Path

from numpy.typing import NDArray

class Transformer:
    def pitch(
        self,
        n_semitones: float,
        quick: bool = False,
    ) -> Self: ...
    def build_array(
        self,
        input_filepath: str | Path | None = None,
        input_array: NDArray[...] | None = None,
        sample_rate_in: float | None = None,
        extra_args: list[str] | None = None,
    ) -> NDArray[...]: ...



from typing import Optional, Tuple

from itertools import chain

import torch
from torch.nn.utils import remove_weight_norm
from torch.nn.utils.parametrizations import weight_norm

from ultimate_rvc.rvc.lib.algorithm.commons import get_padding, init_weights
from ultimate_rvc.rvc.lib.algorithm.modules import WaveNet

LRELU_SLOPE = 0.1


def create_conv1d_layer(channels, kernel_size, dilation):
    return weight_norm(
        torch.nn.Conv1d(
            channels,
            channels,
            kernel_size,
            1,
            dilation=dilation,
            padding=get_padding(kernel_size, dilation),
        ),
    )


def apply_mask(tensor: torch.Tensor, mask: torch.Tensor | None):
    return tensor * mask if mask else tensor


def apply_mask_(tensor: torch.Tensor, mask: torch.Tensor | None):
    return tensor.mul_(mask) if mask else tensor


class ResBlock(torch.nn.Module):
    """
    A residual block module that applies a series of 1D convolutional layers with residual connections.
    """

    def __init__(
        self,
        channels: int,
        kernel_size: int = 3,
        dilations: tuple[int] = (1, 3, 5),
    ):
        """
        Initializes the ResBlock.

        Args:
            channels (int): Number of input and output channels for the convolution layers.
            kernel_size (int): Size of the convolution kernel. Defaults to 3.
            dilations (Tuple[int]): Tuple of dilation rates for the convolution layers in the first set.

        """
        super().__init__()
        # Create convolutional layers with specified dilations and initialize weights
        self.convs1 = self._create_convs(channels, kernel_size, dilations)
        self.convs2 = self._create_convs(channels, kernel_size, [1] * len(dilations))

    @staticmethod
    def _create_convs(channels: int, kernel_size: int, dilations: tuple[int]):
        """
        Creates a list of 1D convolutional layers with specified dilations.

        Args:
            channels (int): Number of input and output channels for the convolution layers.
            kernel_size (int): Size of the convolution kernel.
            dilations (Tuple[int]): Tuple of dilation rates for each convolution layer.

        """
        layers = torch.nn.ModuleList(
            [create_conv1d_layer(channels, kernel_size, d) for d in dilations],
        )
        layers.apply(init_weights)
        return layers

    def forward(self, x: torch.Tensor, x_mask: torch.Tensor = None):
        for conv1, conv2 in zip(self.convs1, self.convs2, strict=False):
            x_residual = x
            x = torch.nn.functional.leaky_relu(x, LRELU_SLOPE)
            x = apply_mask(x, x_mask)
            x = torch.nn.functional.leaky_relu(conv1(x), LRELU_SLOPE)
            x = apply_mask(x, x_mask)
            x = conv2(x)
            x = x + x_residual
        return apply_mask(x, x_mask)

    def remove_weight_norm(self):
        for conv in chain(self.convs1, self.convs2):
            remove_weight_norm(conv)


class Flip(torch.nn.Module):
    """
    Flip module for flow-based models.

    This module flips the input along the time dimension.
    """

    def forward(self, x, *args, reverse=False, **kwargs):
        x = torch.flip(x, [1])
        if not reverse:
            logdet = torch.zeros(x.size(0), dtype=x.dtype, device=x.device)
            return x, logdet
        return x


class ResidualCouplingBlock(torch.nn.Module):
    """
    Residual Coupling Block for normalizing flow.

    Args:
        channels (int): Number of channels in the input.
        hidden_channels (int): Number of hidden channels in the coupling layer.
        kernel_size (int): Kernel size of the convolutional layers.
        dilation_rate (int): Dilation rate of the convolutional layers.
        n_layers (int): Number of layers in the coupling layer.
        n_flows (int, optional): Number of coupling layers in the block. Defaults to 4.
        gin_channels (int, optional): Number of channels for the global conditioning input. Defaults to 0.

    """

    def __init__(
        self,
        channels: int,
        hidden_channels: int,
        kernel_size: int,
        dilation_rate: int,
        n_layers: int,
        n_flows: int = 4,
        gin_channels: int = 0,
    ):
        super(ResidualCouplingBlock, self).__init__()
        self.channels = channels
        self.hidden_channels = hidden_channels
        self.kernel_size = kernel_size
        self.dilation_rate = dilation_rate
        self.n_layers = n_layers
        self.n_flows = n_flows
        self.gin_channels = gin_channels

        self.flows = torch.nn.ModuleList()
        for _ in range(n_flows):
            self.flows.append(
                ResidualCouplingLayer(
                    channels,
                    hidden_channels,
                    kernel_size,
                    dilation_rate,
                    n_layers,
                    gin_channels=gin_channels,
                    mean_only=True,
                ),
            )
            self.flows.append(Flip())

    def forward(
        self,
        x: torch.Tensor,
        x_mask: torch.Tensor,
        g: torch.Tensor | None = None,
        reverse: bool = False,
    ):
        if not reverse:
            for flow in self.flows:
                x, _ = flow(x, x_mask, g=g, reverse=reverse)
        else:
            for flow in reversed(self.flows):
                x = flow.forward(x, x_mask, g=g, reverse=reverse)
        return x

    def remove_weight_norm(self):
        for i in range(self.n_flows):
            self.flows[i * 2].remove_weight_norm()

    def __prepare_scriptable__(self):
        for i in range(self.n_flows):
            for hook in self.flows[i * 2]._forward_pre_hooks.values():
                if (
                    hook.__module__ == "torch.nn.utils.parametrizations.weight_norm"
                    and hook.__class__.__name__ == "WeightNorm"
                ):
                    torch.nn.utils.remove_weight_norm(self.flows[i * 2])

        return self


class ResidualCouplingLayer(torch.nn.Module):
    """
    Residual coupling layer for flow-based models.

    Args:
        channels (int): Number of channels.
        hidden_channels (int): Number of hidden channels.
        kernel_size (int): Size of the convolutional kernel.
        dilation_rate (int): Dilation rate of the convolution.
        n_layers (int): Number of convolutional layers.
        p_dropout (float, optional): Dropout probability. Defaults to 0.
        gin_channels (int, optional): Number of conditioning channels. Defaults to 0.
        mean_only (bool, optional): Whether to use mean-only coupling. Defaults to False.

    """

    def __init__(
        self,
        channels: int,
        hidden_channels: int,
        kernel_size: int,
        dilation_rate: int,
        n_layers: int,
        p_dropout: float = 0,
        gin_channels: int = 0,
        mean_only: bool = False,
    ):
        assert channels % 2 == 0, "channels should be divisible by 2"
        super().__init__()
        self.channels = channels
        self.hidden_channels = hidden_channels
        self.kernel_size = kernel_size
        self.dilation_rate = dilation_rate
        self.n_layers = n_layers
        self.half_channels = channels // 2
        self.mean_only = mean_only

        self.pre = torch.nn.Conv1d(self.half_channels, hidden_channels, 1)
        self.enc = WaveNet(
            hidden_channels,
            kernel_size,
            dilation_rate,
            n_layers,
            p_dropout=p_dropout,
            gin_channels=gin_channels,
        )
        self.post = torch.nn.Conv1d(
            hidden_channels,
            self.half_channels * (2 - mean_only),
            1,
        )
        self.post.weight.data.zero_()
        self.post.bias.data.zero_()

    def forward(
        self,
        x: torch.Tensor,
        x_mask: torch.Tensor,
        g: torch.Tensor | None = None,
        reverse: bool = False,
    ):
        x0, x1 = torch.split(x, [self.half_channels] * 2, 1)
        h = self.pre(x0) * x_mask
        h = self.enc(h, x_mask, g=g)
        stats = self.post(h) * x_mask
        if not self.mean_only:
            m, logs = torch.split(stats, [self.half_channels] * 2, 1)
        else:
            m = stats
            logs = torch.zeros_like(m)

        if not reverse:
            x1 = m + x1 * torch.exp(logs) * x_mask
            x = torch.cat([x0, x1], 1)
            logdet = torch.sum(logs, [1, 2])
            return x, logdet
        x1 = (x1 - m) * torch.exp(-logs) * x_mask
        x = torch.cat([x0, x1], 1)
        return x

    def remove_weight_norm(self):
        self.enc.remove_weight_norm()



from typing import Optional

import numpy as np

import torch
from torch.nn.utils import remove_weight_norm
from torch.nn.utils.parametrizations import weight_norm

from ultimate_rvc.rvc.lib.algorithm.commons import init_weights
from ultimate_rvc.rvc.lib.algorithm.residuals import LRELU_SLOPE, ResBlock


class HiFiGANGenerator(torch.nn.Module):
    """
    HiFi-GAN Generator module for audio synthesis.

    This module implements the generator part of the HiFi-GAN architecture,
    which uses transposed convolutions for upsampling and residual blocks for
    refining the audio output. It can also incorporate global conditioning.

    Args:
        initial_channel (int): Number of input channels to the initial convolutional layer.
        resblock_kernel_sizes (list): List of kernel sizes for the residual blocks.
        resblock_dilation_sizes (list): List of lists of dilation rates for the residual blocks, corresponding to each kernel size.
        upsample_rates (list): List of upsampling factors for each upsampling layer.
        upsample_initial_channel (int): Number of output channels from the initial convolutional layer, which is also the input to the first upsampling layer.
        upsample_kernel_sizes (list): List of kernel sizes for the transposed convolutional layers used for upsampling.
        gin_channels (int, optional): Number of input channels for the global conditioning. If 0, no global conditioning is used. Defaults to 0.

    """

    def __init__(
        self,
        initial_channel: int,
        resblock_kernel_sizes: list,
        resblock_dilation_sizes: list,
        upsample_rates: list,
        upsample_initial_channel: int,
        upsample_kernel_sizes: list,
        gin_channels: int = 0,
    ):
        super(HiFiGANGenerator, self).__init__()
        self.num_kernels = len(resblock_kernel_sizes)
        self.num_upsamples = len(upsample_rates)
        self.conv_pre = torch.nn.Conv1d(
            initial_channel,
            upsample_initial_channel,
            7,
            1,
            padding=3,
        )

        self.ups = torch.nn.ModuleList()
        self.resblocks = torch.nn.ModuleList()

        for i, (u, k) in enumerate(
            zip(upsample_rates, upsample_kernel_sizes, strict=False),
        ):
            self.ups.append(
                weight_norm(
                    torch.nn.ConvTranspose1d(
                        upsample_initial_channel // (2**i),
                        upsample_initial_channel // (2 ** (i + 1)),
                        k,
                        u,
                        padding=(k - u) // 2,
                    ),
                ),
            )
            ch = upsample_initial_channel // (2 ** (i + 1))
            for j, (k, d) in enumerate(
                zip(resblock_kernel_sizes, resblock_dilation_sizes, strict=False),
            ):
                self.resblocks.append(ResBlock(ch, k, d))

        self.conv_post = torch.nn.Conv1d(ch, 1, 7, 1, padding=3, bias=False)
        self.ups.apply(init_weights)

        if gin_channels != 0:
            self.cond = torch.nn.Conv1d(gin_channels, upsample_initial_channel, 1)

    def forward(self, x: torch.Tensor, g: torch.Tensor | None = None):
        # new tensor
        x = self.conv_pre(x)

        if g is not None:
            x = x + self.cond(g)

        for i in range(self.num_upsamples):
            x = torch.nn.functional.leaky_relu(x, LRELU_SLOPE)
            x = self.ups[i](x)
            xs = None
            for j in range(self.num_kernels):
                if xs is None:
                    xs = self.resblocks[i * self.num_kernels + j](x)
                else:
                    xs += self.resblocks[i * self.num_kernels + j](x)
            x = xs / self.num_kernels
        # in-place call
        x = torch.nn.functional.leaky_relu(x)
        x = self.conv_post(x)
        # in-place call
        x = torch.tanh(x)

        return x

    def __prepare_scriptable__(self):
        for l in self.ups_and_resblocks:
            for hook in l._forward_pre_hooks.values():
                if (
                    hook.__module__ == "torch.nn.utils.parametrizations.weight_norm"
                    and hook.__class__.__name__ == "WeightNorm"
                ):
                    torch.nn.utils.remove_weight_norm(l)
        return self

    def remove_weight_norm(self):
        for l in self.ups:
            remove_weight_norm(l)
        for l in self.resblocks:
            l.remove_weight_norm()


class SineGenerator(torch.nn.Module):
    """
    Sine wave generator with optional harmonic overtones and noise.

    This module generates sine waves for a fundamental frequency and its harmonics.
    It can also add Gaussian noise and apply a voiced/unvoiced mask.

    Args:
        sampling_rate (int): The sampling rate of the audio in Hz.
        num_harmonics (int, optional): The number of harmonic overtones to generate. Defaults to 0.
        sine_amplitude (float, optional): The amplitude of the sine wave components. Defaults to 0.1.
        noise_stddev (float, optional): The standard deviation of the additive Gaussian noise. Defaults to 0.003.
        voiced_threshold (float, optional): The threshold for the fundamental frequency (F0) to determine if a frame is voiced. Defaults to 0.0.

    """

    def __init__(
        self,
        sampling_rate: int,
        num_harmonics: int = 0,
        sine_amplitude: float = 0.1,
        noise_stddev: float = 0.003,
        voiced_threshold: float = 0.0,
    ):
        super(SineGenerator, self).__init__()
        self.sampling_rate = sampling_rate
        self.num_harmonics = num_harmonics
        self.sine_amplitude = sine_amplitude
        self.noise_stddev = noise_stddev
        self.voiced_threshold = voiced_threshold
        self.waveform_dim = self.num_harmonics + 1  # fundamental + harmonics

    def _compute_voiced_unvoiced(self, f0: torch.Tensor):
        """
        Generates a binary mask indicating voiced/unvoiced frames based on the fundamental frequency.

        Args:
            f0 (torch.Tensor): Fundamental frequency tensor of shape (batch_size, length).

        """
        uv_mask = (f0 > self.voiced_threshold).float()
        return uv_mask

    def _generate_sine_wave(self, f0: torch.Tensor, upsampling_factor: int):
        """
        Generates sine waves for the fundamental frequency and its harmonics.

        Args:
            f0 (torch.Tensor): Fundamental frequency tensor of shape (batch_size, length, 1).
            upsampling_factor (int): The factor by which to upsample the sine wave.

        """
        batch_size, length, _ = f0.shape

        # Create an upsampling grid
        upsampling_grid = torch.arange(
            1,
            upsampling_factor + 1,
            dtype=f0.dtype,
            device=f0.device,
        )

        # Calculate phase increments
        phase_increments = (f0 / self.sampling_rate) * upsampling_grid
        phase_remainder = torch.fmod(phase_increments[:, :-1, -1:] + 0.5, 1.0) - 0.5
        cumulative_phase = phase_remainder.cumsum(dim=1).fmod(1.0).to(f0.dtype)
        phase_increments += torch.nn.functional.pad(
            cumulative_phase,
            (0, 0, 1, 0),
            mode="constant",
        )

        # Reshape to match the sine wave shape
        phase_increments = phase_increments.reshape(batch_size, -1, 1)

        # Scale for harmonics
        harmonic_scale = torch.arange(
            1,
            self.waveform_dim + 1,
            dtype=f0.dtype,
            device=f0.device,
        ).reshape(1, 1, -1)
        phase_increments *= harmonic_scale

        # Add random phase offset (except for the fundamental)
        random_phase = torch.rand(1, 1, self.waveform_dim, device=f0.device)
        random_phase[..., 0] = 0  # Fundamental frequency has no random offset
        phase_increments += random_phase

        # Generate sine waves
        sine_waves = torch.sin(2 * np.pi * phase_increments)
        return sine_waves

    def forward(self, f0: torch.Tensor, upsampling_factor: int):
        with torch.no_grad():
            # Expand `f0` to include waveform dimensions
            f0 = f0.unsqueeze(-1)

            # Generate sine waves
            sine_waves = (
                self._generate_sine_wave(f0, upsampling_factor) * self.sine_amplitude
            )

            # Compute voiced/unvoiced mask
            voiced_mask = self._compute_voiced_unvoiced(f0)

            # Upsample voiced/unvoiced mask
            voiced_mask = torch.nn.functional.interpolate(
                voiced_mask.transpose(2, 1),
                scale_factor=float(upsampling_factor),
                mode="nearest",
            ).transpose(2, 1)

            # Compute noise amplitude
            noise_amplitude = voiced_mask * self.noise_stddev + (1 - voiced_mask) * (
                self.sine_amplitude / 3
            )

            # Add Gaussian noise
            noise = noise_amplitude * torch.randn_like(sine_waves)

            # Combine sine waves and noise
            sine_waveforms = sine_waves * voiced_mask + noise

        return sine_waveforms, voiced_mask, noise



import asyncio
import os
import sys

import edge_tts


async def main():
    # Parse command line arguments
    tts_file = str(sys.argv[1])
    text = str(sys.argv[2])
    voice = str(sys.argv[3])
    rate = int(sys.argv[4])
    output_file = str(sys.argv[5])

    rates = f"+{rate}%" if rate >= 0 else f"{rate}%"
    if tts_file and os.path.exists(tts_file):
        text = ""
        try:
            with open(tts_file, encoding="utf-8") as file:
                text = file.read()
        except UnicodeDecodeError:
            with open(tts_file) as file:
                text = file.read()
    await edge_tts.Communicate(text, voice, rate=rates).save(output_file)
    # print(f"TTS with {voice} completed. Output TTS file: '{output_file}'")


if __name__ == "__main__":
    asyncio.run(main())



import json
import os

import torch

from ultimate_rvc.rvc.common import RVC_CONFIGS_DIR

version_config_paths = [
    os.path.join("48000.json"),
    os.path.join("40000.json"),
    os.path.join("32000.json"),
]


def singleton(cls):
    instances = {}

    def get_instance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return get_instance


@singleton
class Config:
    def __init__(self):
        self.device = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.gpu_name = (
            torch.cuda.get_device_name(int(self.device.split(":")[-1]))
            if self.device.startswith("cuda")
            else None
        )
        self.json_config = self.load_config_json()
        self.gpu_mem = None
        self.x_pad, self.x_query, self.x_center, self.x_max = self.device_config()

    def load_config_json(self) -> dict:
        configs = {}
        for config_file in version_config_paths:
            config_path = os.path.join(str(RVC_CONFIGS_DIR), config_file)
            with open(config_path) as f:
                configs[config_file] = json.load(f)
        return configs

    def has_mps(self) -> bool:
        # Check if Metal Performance Shaders are available - for macOS 12.3+.
        return torch.backends.mps.is_available()

    def has_xpu(self) -> bool:
        # Check if XPU is available.
        return hasattr(torch, "xpu") and torch.xpu.is_available()

    def device_config(self) -> tuple:
        if self.device.startswith("cuda"):
            self.set_cuda_config()
        elif self.has_mps():
            self.device = "mps"
        else:
            self.device = "cpu"

        # Configuration for 6GB GPU memory
        x_pad, x_query, x_center, x_max = (1, 6, 38, 41)
        if self.gpu_mem is not None and self.gpu_mem <= 4:
            # Configuration for 5GB GPU memory
            x_pad, x_query, x_center, x_max = (1, 5, 30, 32)

        return x_pad, x_query, x_center, x_max

    def set_cuda_config(self):
        i_device = int(self.device.split(":")[-1])
        self.gpu_name = torch.cuda.get_device_name(i_device)

        self.gpu_mem = torch.cuda.get_device_properties(i_device).total_memory // (
            1024**3
        )


def max_vram_gpu(gpu):
    if torch.cuda.is_available():
        gpu_properties = torch.cuda.get_device_properties(gpu)
        total_memory_gb = round(gpu_properties.total_memory / 1024 / 1024 / 1024)
        return total_memory_gb
    return "8"


def get_gpu_info():
    ngpu = torch.cuda.device_count()
    gpu_infos = []
    if torch.cuda.is_available() or ngpu != 0:
        for i in range(ngpu):
            gpu_name = torch.cuda.get_device_name(i)
            mem = int(
                torch.cuda.get_device_properties(i).total_memory / 1024 / 1024 / 1024
                + 0.4,
            )
            gpu_infos.append(f"{i}: {gpu_name} ({mem} GB)")
    if len(gpu_infos) > 0:
        gpu_info = "\n".join(gpu_infos)
    else:
        gpu_info = (
            "Unfortunately, there is no compatible GPU available to support your"
            " training."
        )
    return gpu_info


def get_number_of_gpus():
    if torch.cuda.is_available():
        num_gpus = torch.cuda.device_count()
        return "-".join(map(str, range(num_gpus)))
    return "-"



"""
Module which defines custom exception and enumerations used when
instiating and re-raising those exceptions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from enum import StrEnum

if TYPE_CHECKING:
    from ultimate_rvc.typing_extra import PretrainedSampleRate, StrPath


class Entity(StrEnum):
    """Enumeration of entities that can be provided."""

    # General entities
    FILE = "file"
    FILES = "files"
    DIRECTORY = "directory"
    DIRECTORIES = "directories"

    # Model entities
    MODEL_NAME = "model name"
    MODEL_NAMES = "model names"
    UPLOAD_NAME = "upload name"
    MODEL = "model"
    VOICE_MODEL = "voice model"
    TRAINING_MODEL = "training model"
    CUSTOM_EMBEDDER_MODEL = "custom embedder model"
    CUSTOM_PRETRAINED_MODEL = "custom pretrained model"
    GENERATOR = "generator"
    DISCRIMINATOR = "discriminator"
    MODEL_FILE = "model file"
    MODEL_BIN_FILE = "pytorch_model.bin file"
    CONFIG_JSON_FILE = "config.json file"

    # Audio entities
    AUDIO_TRACK = "audio track"
    AUDIO_TRACK_GAIN_PAIRS = "pairs of audio track and gain"
    VOICE_TRACK = "voice track"
    SPEECH_TRACK = "speech track"
    VOCALS_TRACK = "vocals track"
    SONG_DIR = "song directory"
    DATASET = "dataset"
    DATASETS = "datasets"
    DATASET_NAME = "dataset name"
    DATASET_FILE_LIST = "dataset file list"
    PREPROCESSED_AUDIO_DATASET_FILES = "preprocessed dataset audio files"

    # Source entities
    SOURCE = "source"
    URL = "URL"

    # GPU entities
    GPU_IDS = "GPU IDs"

    # Config entities
    CONFIG = "configuration"
    CONFIG_NAME = "configuration name"
    CONFIG_NAMES = "configuration names"
    EVENT = "event"
    COMPONENT = "component"


AudioFileEntity = Literal[
    Entity.AUDIO_TRACK,
    Entity.VOICE_TRACK,
    Entity.SPEECH_TRACK,
    Entity.VOCALS_TRACK,
    Entity.FILE,
]

AudioDirectoryEntity = Literal[Entity.SONG_DIR, Entity.DATASET, Entity.DIRECTORY]

ModelEntity = Literal[
    Entity.MODEL,
    Entity.VOICE_MODEL,
    Entity.TRAINING_MODEL,
    Entity.CUSTOM_EMBEDDER_MODEL,
    Entity.CUSTOM_PRETRAINED_MODEL,
]

ConfigEntity = Literal[Entity.EVENT, Entity.COMPONENT]


class Location(StrEnum):
    """Enumeration of locations where entities can be found."""

    # Audio locations
    AUDIO_ROOT = "the root of the audio base directory"
    INTERMEDIATE_AUDIO_ROOT = "the root of the intermediate audio base directory"
    SPEECH_AUDIO_ROOT = "the root of the speech audio directory"
    TRAINING_AUDIO_ROOT = "the root of the training audio directory"
    OUTPUT_AUDIO_ROOT = "the root of the output audio directory"

    # Model locations
    EXTRACTED_ZIP_FILE = "extracted zip file"


class Step(StrEnum):
    """Enumeration of steps that can be run."""

    DATASET_PREPROCESSING = "dataset preprocessing"
    FEATURE_EXTRACTION = "feature extraction"


class UIMessage(StrEnum):
    """
    Enumeration of messages that can be displayed in the UI
    in place of core exception messages.
    """

    # General messages
    NO_UPLOADED_FILES = "No files selected."

    # Audio messages
    NO_AUDIO_TRACK = "No audio tracks provided."
    NO_SPEECH_AUDIO_FILES = (
        "No files selected. Please select one or more speech audio files to delete."
    )
    NO_OUTPUT_AUDIO_FILES = (
        "No files selected. Please select one or more output audio files to delete."
    )
    NO_SONG_DIR = "No song directory selected."
    NO_SONG_DIRS = (
        "No song directories selected. Please select one or more song directories"
        " containing intermediate audio files to delete."
    )
    NO_DATASETS = (
        "No datasets selected. Please select one or more datasets containing audio"
        " files to delete."
    )

    # Model messages
    NO_MODEL = "No model selected."
    NO_MODELS = "No models selected."
    NO_VOICE_MODEL = "No voice model selected."
    NO_VOICE_MODELS = "No voice models selected."
    NO_TRAINING_MODELS = "No training models selected."
    NO_CUSTOM_EMBEDDER_MODEL = "No custom embedder model selected."
    NO_CUSTOM_EMBEDDER_MODELS = "No custom embedder models selected."
    NO_CUSTOM_PRETRAINED_MODELS = "No custom pretrained models selected."
    NO_CUSTOM_PRETRAINED_MODEL = "No custom pretrained model selected."

    # Source messages
    NO_AUDIO_SOURCE = (
        "No source provided. Please provide a valid Youtube URL, local audio file"
        " or song directory."
    )
    NO_TEXT_SOURCE = (
        "No source provided. Please provide a valid text string or path to a text file."
    )

    # GPU messages
    NO_GPUS = "No GPUs selected."

    # Config messages
    NO_CONFIG = "No configuration selected."
    NO_CONFIGS = "No configurations selected."


class NotProvidedError(ValueError):
    """Raised when an entity is not provided."""

    def __init__(self, entity: Entity, ui_msg: UIMessage | None = None) -> None:
        """
        Initialize a NotProvidedError instance.

        Exception message will be formatted as:

        "No `<entity>` provided."

        Parameters
        ----------
        entity : Entity
            The entity that was not provided.
        ui_msg : UIMessage, default=None
            Message which, if provided, is displayed in the UI
            instead of the default exception message.

        """
        super().__init__(f"No {entity} provided.")
        self.ui_msg = ui_msg


class NotFoundError(OSError):
    """Raised when an entity is not found in a given location."""

    def __init__(
        self,
        entity: Entity,
        location: StrPath | Location,
        is_path: bool = True,
    ) -> None:
        """
        Initialize a NotFoundError instance.

        Exception message will be formatted as:

        "`<entity>` not found `(`in `|` at:`)` `<location>`."

        Parameters
        ----------
        entity : Entity
            The entity that was not found.
        location : StrPath | Location
            The location where the entity was not found.
        is_path : bool, default=True
            Whether the location is a path to the entity.

        """
        proposition = "at:" if is_path else "in"
        entity_cap = entity.capitalize() if not entity.isupper() else entity
        super().__init__(
            f"{entity_cap} not found {proposition} {location}",
        )


class EntityNotFoundError(OSError):
    """Raised when an entity is not found."""

    def __init__(self, entity: Entity, name: str) -> None:
        r"""
        Initialize an EntityNotFoundError instance.

        Exception message will be formatted as:

        "`<entity>` with name '`<name>`' not found."

        Parameters
        ----------
        entity : Entity
            The entity that was not found.
        name : str
            The name of the entity that was not found.

        """
        super().__init__(f"{entity.capitalize()} with name '{name}' not found.")


class ModelNotFoundError(EntityNotFoundError):
    """Raised when an model with a given name is not found."""

    def __init__(self, entity: ModelEntity, name: str) -> None:
        r"""
        Initialize a ModelNotFoundError instance.

        Exception message will be formatted as:

        '`<entity>` with name "`<name>`" not found.'

        Parameters
        ----------
        entity : ModelEntity
            The model entity that was not found.
        name : str
            The name of the model that was not found.

        """
        super().__init__(entity, name)


class ConfigNotFoundError(EntityNotFoundError):
    """Raised when a configuration with a given name is not found."""

    def __init__(self, name: str) -> None:
        r"""
        Initialize a ConfigNotFoundError instance.

        Exception message will be formatted as:

        'Configuration with name '`<name>`' not found.'

        Parameters
        ----------
        name : str
            The name of the configuration that was not found.

        """
        super().__init__(Entity.CONFIG, name)


class PretrainedModelNotAvailableError(OSError):
    """Raised when a pretrained model is not available for download."""

    def __init__(
        self,
        name: str,
        sample_rate: PretrainedSampleRate | None = None,
    ) -> None:
        r"""
        Initialize a PretrainedModelNotAvailableError instance.

        Exception message will be formatted as:

        'Pretrained model with name "`<name>`"
        [and sample rate `<sample_rate>`] is not available for
        download.'

        Parameters
        ----------
        name : str
            The name of the pretrained model that is not available for
            download.
        sample_rate : PretrainedSampleRate, optional
            The sample rate of the pretrained model that is not
            available for download.

        """
        suffix = f" and sample rate {sample_rate}" if sample_rate else ""
        super().__init__(
            f"Pretrained model with name '{name}'{suffix} is not available for"
            " download.",
        )


class IncompatiblePretrainedModelError(OSError):
    """
    Raised when a pretrained model is incompatible with a given sample
    rate.
    """

    def __init__(self, name: str, sample_rate: int) -> None:
        r"""
        Initialize an IncompatiblePretrainedModelError instance.

        Exception message will be formatted as:

        'Pretrained model with name "`<name>`" is incompatible with
        sample rate `<sample_rate>`.'

        Parameters
        ----------
        name : str
            The name of the pretrained model that is incompatible with
            a given sample rate.
        sample_rate : int
            The sample rate that the pretrained model is incompatible
            with.

        """
        super().__init__(
            f"Pretrained model with name '{name}' is incompatible with sample rate"
            f" {sample_rate}.",
        )


class IncompatibleVocoderError(OSError):
    """
    Raised when a default pretrained model is incompatible with a
    given vocoder.
    """

    def __init__(self, vocoder: str) -> None:
        r"""
        Initialize an IncompatibleVocoderError instance.

        Exception message will be formatted as:

        'The default pretrained model is incompatible with the vocoder
        `<vocoder>`.'

        Parameters
        ----------
        vocoder : str
            The vocoder that the default pretrained model is
            incompatible with.

        """
        super().__init__(
            f"The default pretrained model is incompatible with the vocoder {vocoder}.",
        )


class GPUNotFoundError(OSError):
    """Raised when a GPU with a given id is not found."""

    def __init__(self, device_id: int | None = None) -> None:
        r"""
        Initialize a GPUNotFoundError instance.

        Exception message will be formatted as:

        'GPU with id `<id>` not found.'

        Parameters
        ----------
        device_id : int, optional
            The id of a GPU that is not found.

        """
        super().__init__(
            f"No GPU with id {device_id} found.",
        )


class ModelAsssociatedEntityNotFoundError(OSError):
    """Raised when an entity associated with a model is not found."""

    def __init__(
        self,
        entity: Entity,
        model_name: str,
        required_step: Step | None = None,
    ) -> None:
        r"""
        Initialize a ModelAsssociatedEntityNotFoundError instance.

        Exception message will be formatted as:

        'No `<entity>` associated with the model with name
        "`<model_name>`". [Please run `<required_step>` first.]'

        Parameters
        ----------
        entity : Entity
            The entity that is not associated with the model.
        model_name : str
            The name of the model that the entity is not associated
            with.
        required_step : Step | None, default=None
            The required step that needs to be run before will be
            associated with the model.

        """
        suffix = f"Please run {required_step} first." if required_step else ""
        super().__init__(
            f"No {entity.capitalize()} associated with the model with name"
            f" {model_name}. {suffix}",
        )


class EntityExistsError(OSError):
    """Raised when an entity already exists."""

    def __init__(self, entity: Entity, name: str) -> None:
        r"""
        Initialize an EntityExistsError instance.

        Exception message will be formatted as:

        '`<entity>` with name '`<name>`' already exists. Please provide
        a different name for your {entity}.'

        Parameters
        ----------
        entity : Entity
            The entity that already exists.
        name : str
            The name of the entity that already exists.

        """
        super().__init__(
            f"{entity.capitalize()} with name '{name}' already exists. Please provide a"
            f" different name for your {entity}.",
        )


class ModelExistsError(EntityExistsError):
    """Raised when a model already exists."""

    def __init__(self, entity: ModelEntity, name: str) -> None:
        r"""
        Initialize a ModelExistsError instance.

        Exception message will be formatted as:

        '`<entity>` with name "`<name>`" already exists. Please provide
        a different name for your {entity}.'

        Parameters
        ----------
        entity : ModelEntity
            The model entity that already exists.
        name : str
            The name of the model that already exists.

        """
        super().__init__(entity, name)


class ConfigExistsError(EntityExistsError):
    """Raised when a configuration already exists."""

    def __init__(self, name: str) -> None:
        r"""
        Initialize a ConfigExistsError instance.

        Exception message will be formatted as:

        'Configuration with name '`<name>`' already exists. Please
        provide a different name for your {entity}.'

        Parameters
        ----------
        name : str
            The name of the configuration that already exists.

        """
        super().__init__(Entity.CONFIG, name)


class PretrainedModelExistsError(OSError):
    """
    Raised when a pretrained model with a given name and sample rate
    already exists.
    """

    def __init__(self, name: str, sample_rate: PretrainedSampleRate) -> None:
        r"""
        Initialize a PretrainedModelExistsError instance.

        Exception message will be formatted as:

        'Pretrained model with name "`<name>`" and sample rate
        `<sample_rate>` already exists.'

        Parameters
        ----------
        name : str
            The name of the pretrained model that already exists.
        sample_rate : PretrainedSampleRate
            The sample rate of the pretrained model that already exists.

        """
        super().__init__(
            f"Pretrained model with name '{name}' and sample rate {sample_rate} already"
            " exists.",
        )


class InvalidLocationError(OSError):
    """Raised when an entity is in a wrong location."""

    def __init__(self, entity: Entity, location: Location, path: StrPath) -> None:
        r"""
        Initialize an InvalidLocationError instance.

        Exception message will be formatted as:

        "`<entity>` should be located in `<location>` but found at:
        `<path>`"

        Parameters
        ----------
        entity : Entity
            The entity that is in a wrong location.
        location : Location
            The correct location for the entity.
        path : StrPath
            The path to the entity.

        """
        entity_cap = entity.capitalize() if not entity.isupper() else entity
        super().__init__(
            f"{entity_cap} should be located in {location} but found at: {path}",
        )


class HttpUrlError(OSError):
    """Raised when a HTTP-based URL is invalid."""

    def __init__(self, url: str) -> None:
        """
        Initialize a HttpUrlError instance.

        Exception message will be formatted as:

        "Invalid HTTP-based URL: `<url>`"

        Parameters
        ----------
        url : str
            The invalid HTTP-based URL.

        """
        super().__init__(
            f"Invalid HTTP-based URL: {url}",
        )


class YoutubeUrlError(OSError):
    """
    Raised when an URL does not point to a YouTube video or
    , potentially, a Youtube playlist.
    """

    def __init__(self, url: str, playlist: bool) -> None:
        """
        Initialize a YoutubeURlError instance.

        Exception message will be formatted as:

        "URL does not point to a YouTube video `[`or playlist`]`:
         `<url>`"

        Parameters
        ----------
        url : str
            The URL that does not point to a YouTube video or playlist.
        playlist : bool
            Whether the URL might point to a YouTube playlist.

        """
        suffix = " or playlist" if playlist else ""
        super().__init__(
            f"Not able to access Youtube video{suffix} at: {url}",
        )


class UploadLimitError(ValueError):
    """Raised when the upload limit for an entity is exceeded."""

    def __init__(self, entity: Entity, limit: str | float) -> None:
        """
        Initialize an UploadLimitError instance.

        Exception message will be formatted as:

        "At most `<limit>` `<entity>` can be uploaded."

        Parameters
        ----------
        entity : Entity
            The entity for which the upload limit was exceeded.
        limit : str
            The upload limit.

        """
        super().__init__(f"At most {limit} {entity} can be uploaded.")


class UploadTypeError(ValueError):
    """
    Raised when one or more uploaded entities have an invalid
    type.
    """

    def __init__(
        self,
        entity: Entity,
        valid_types: list[str],
        type_class: Literal["formats", "names"],
        multiple: bool,
    ) -> None:
        """
        Initialize an UploadTypeError instance.

        Exception message will be formatted as:

        "Only `<entity>` with the following `<type_class>` can be
        uploaded `(`by themselves | together`)`: `<types>`."

        Parameters
        ----------
        entity : Entity
            The entity with an invalid type that was uploaded.
        valid_types : list[str]
            The valid types for the entity that was uploaded.
        type_class : Literal["formats", "names"]
            The name for the class of valid types.
        multiple : bool
            Whether multiple instances of the entity were uploaded.

        """
        suffix = "by themselves" if not multiple else "together (at most one of each)"
        super().__init__(
            f"Only {entity} with the following {type_class} can be uploaded {suffix}:"
            f" {', '.join(valid_types)}.",
        )


class InvalidAudioFormatError(ValueError):
    """Raised when an audio file has an invalid format."""

    def __init__(self, path: StrPath, formats: list[str]) -> None:
        """
        Initialize an InvalidAudioFormatError instance.

        Exception message will be formatted as:

        "Invalid audio file format: `<path>`. Supported formats are:
        `<formats>`."

        Parameters
        ----------
        path : StrPath
            The path to the audio file with an invalid format.
        formats : list[str]
            Supported audio formats.

        """
        super().__init__(
            f"Invalid audio file format: {path}. Supported formats are:"
            f" {', '.join(formats)}.",
        )


class NotInstantiatedError(ValueError):
    """Raised when an entity is not instantiated."""

    def __init__(self, entity: Entity) -> None:
        """
        Initialize a NotInstantiatedError instance.

        Exception message will be formatted as:

        "`<entity>` has not been instantiated."

        """
        super().__init__(f"{entity} has not been instantiated.")


class ComponentNotInstatiatedError(NotInstantiatedError):
    """Raised when a component is not instantiated."""

    def __init__(self) -> None:
        """
        Initialize a ComponentNotInstantiatedError instance.

        Exception message will be formatted as:

        "Component has not been instantiated."

        """
        super().__init__(Entity.COMPONENT)


class EventNotInstantiatedError(NotInstantiatedError):
    """Raised when an event is not instantiated."""

    def __init__(self) -> None:
        """
        Initialize a EventNotInstantiatedError instance.

        Exception message will be formatted as:

        "Event has not been instantiated."

        """
        super().__init__(Entity.EVENT)



"""
Common definitions for modules in the Ultimate RVC project that
facilitate management.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import operator
import shutil
from pathlib import Path

if TYPE_CHECKING:
    from ultimate_rvc.typing_extra import StrPath


def get_named_items(
    directory: StrPath,
    exclude: str | None = None,
    include_suffix: bool = True,
) -> list[tuple[str, str]]:
    """
    Get the names and paths of all items in the provided directory
    sorted by name.

    Parameters
    ----------
    directory : str
        The path to the directory containing the items.
    exclude : str, optional
        The file extension to exclude from the list of items.
    include_suffix : bool, optional
        Whether to include the file extension in the names of the items.

    Returns
    -------
    list[tuple[str, str]]
        A list of tuples containing the names and paths of all items in
        the directory sorted by name.

    """
    dir_path = Path(directory)
    if dir_path.is_dir():
        named_items = sorted(
            [
                (item_path.name if include_suffix else item_path.stem, str(item_path))
                for item_path in dir_path.iterdir()
                if item_path.suffix != exclude
            ],
        )
        return sorted(named_items, key=operator.itemgetter(0))
    return []


def get_items(
    directory: StrPath,
    only_stem: bool = True,
    exclude: str | None = None,
) -> list[str]:
    """
    Get the names or paths of all items in the provided directory.

    Parameters
    ----------
    directory : str
        The path to the directory containing the items.
    only_stem : bool, optional
        Whether to return only the stem of the items or their paths.
        For directories, returns the full name. For files, returns the
        name without the file extension.
    exclude : str, optional
        The file extension to exclude from the list of items.

    Returns
    -------
    list[str]
        A list of the names or paths of all items in the directory.

    """
    dir_path = Path(directory)
    if dir_path.is_dir():
        return sorted(
            [
                (
                    (item_path.name if item_path.is_dir() else item_path.stem)
                    if only_stem
                    else str(item_path)
                )
                for item_path in dir_path.iterdir()
                if item_path.suffix != exclude
            ],
        )
    return []


def delete_directory(path: StrPath) -> None:
    """
    Delete a directory and its contents if it exists.

    Parameters
    ----------
    path : str
        The path to the directory to delete.

    """
    dir_path = Path(path)
    if dir_path.is_dir():
        shutil.rmtree(dir_path)



"""
Module which exposes functionality for extracting training features from
audio datasets.
"""

from __future__ import annotations

from multiprocessing import cpu_count

from ultimate_rvc.core.common import (
    display_progress,
    get_combined_file_hash,
    validate_model,
)
from ultimate_rvc.core.exceptions import (
    Entity,
    ModelAsssociatedEntityNotFoundError,
    Step,
)
from ultimate_rvc.core.train.common import validate_devices
from ultimate_rvc.typing_extra import (
    DeviceType,
    EmbedderModel,
    TrainingF0Method,
)


def extract_features(
    model_name: str,
    f0_method: TrainingF0Method = TrainingF0Method.RMVPE,
    hop_length: int = 128,
    embedder_model: EmbedderModel = EmbedderModel.CONTENTVEC,
    custom_embedder_model: str | None = None,
    include_mutes: int = 2,
    cpu_cores: int = cpu_count(),
    hardware_acceleration: DeviceType = DeviceType.AUTOMATIC,
    gpu_ids: set[int] | None = None,
) -> None:
    """
    Extract features from the preprocessed dataset associated with a
    voice model to be trained.

    Parameters
    ----------
    model_name : str
        The name of the voice model to be trained.
    f0_method : TrainingF0Method, default=TrainingF0Method.RMVPE
        The method to use for extracting pitch features.
    hop_length : int, default=128
        The hop length to use for extracting pitch features. Only used
        with the CREPE pitch extraction method.
    embedder_model : EmbedderModel, default=EmbedderModel.CONTENTVEC
        The model to use for extracting audio embeddings.
    custom_embedder_model : StrPath, optional
        The name of the custom embedder model to use for extracting
        audio embeddings.
    include_mutes : int, default=2
        The number of mute audio files to include in the generated
        training file list. Adding silent files enables the voice model
        to handle pure silence in inferred audio files. If the
        preprocessed audio dataset already contains segments of pure
        silence, set this to 0.
    cpu_cores : int, default=cpu_count()
        The number of CPU cores to use for feature extraction.
    hardware_acceleration : DeviceType, default=DeviceType.AUTOMATIC
        The type of hardware acceleration to use for feature extraction.
        `AUTOMATIC` will select the first available GPU and fall back to
        CPU if no GPUs are available.
    gpu_ids : set[int], optional
        Set of ids of the GPUs to use for feature extraction when `GPU`
        is selected for hardware acceleration.

    Raises
    ------
    ModelAsssociatedEntityNotFoundError
        If no preprocessed dataset audio files are associated with the
        voice model identified by the provided name.

    """
    model_path = validate_model(model_name, Entity.TRAINING_MODEL)
    sliced_audios16k_path = model_path / "sliced_audios_16k"
    if not sliced_audios16k_path.is_dir() or not any(sliced_audios16k_path.iterdir()):
        raise ModelAsssociatedEntityNotFoundError(
            Entity.PREPROCESSED_AUDIO_DATASET_FILES,
            model_name,
            Step.DATASET_PREPROCESSING,
        )

    custom_embedder_model_path, combined_file_hash = None, None
    chosen_embedder_model, embedder_model_id = [embedder_model] * 2
    if embedder_model == EmbedderModel.CUSTOM:
        custom_embedder_model_path = validate_model(
            custom_embedder_model,
            Entity.CUSTOM_EMBEDDER_MODEL,
        )
        json_file = custom_embedder_model_path / "config.json"
        bin_path = custom_embedder_model_path / "pytorch_model.bin"

        combined_file_hash = get_combined_file_hash([json_file, bin_path])
        chosen_embedder_model = str(custom_embedder_model_path)
        embedder_model_id = f"custom_{combined_file_hash}"

    f0_method_id = f0_method
    if f0_method in {TrainingF0Method.CREPE, TrainingF0Method.CREPE_TINY}:
        f0_method_id = f"{f0_method}_{hop_length}"

    device_type, device_ids = validate_devices(hardware_acceleration, gpu_ids)

    devices = (
        [f"{device_type}:{device_id}" for device_id in device_ids]
        if device_ids
        else [device_type]
    )
    # NOTE The lazy_import function does not work with the package below
    # so we import it here manually
    from ultimate_rvc.rvc.train.extract import extract  # noqa: PLC0415

    file_infos = extract.initialize_extraction(
        str(model_path),
        f0_method_id,
        embedder_model_id,
    )
    extract.update_model_info(
        str(model_path),
        chosen_embedder_model,
        combined_file_hash,
    )
    display_progress("[~] Extracting pitch features...")
    extract.run_pitch_extraction(file_infos, devices, f0_method, hop_length, cpu_cores)
    display_progress("[~] Extracting audio embeddings...")
    extract.run_embedding_extraction(
        file_infos,
        devices,
        embedder_model,
        (
            str(custom_embedder_model_path)
            if custom_embedder_model_path is not None
            else None
        ),
        cpu_cores,
    )
    # NOTE The lazy_import function does not work with the package below
    # so we import it here manually
    from ultimate_rvc.rvc.train.extract import preparing_files  # noqa: PLC0415

    preparing_files.generate_config(str(model_path))
    preparing_files.generate_filelist(
        str(model_path),
        include_mutes,
        f0_method_id,
        embedder_model_id,
    )



"""
Module which defines extra types used by modules in the
ultimate_rvc.core.train package.
"""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class ModelInfo(BaseModel):
    """
    Information about a voice model to be trained.

    Attributes
    ----------
    sample_rate : PretrainedSampleRate
        The sample rate of the post-processed audio to train the model
        on.

    """

    sample_rate: int
    # TODO add more attributes later


class TrainingInfo(BaseModel):
    """
    Information about the ongoing training of a voice model.

    Attributes
    ----------
    process_pids : list[int], default = []
        The ids of the processes running the training.

    """

    process_pids: list[int] = []
    # TODO add more attributes later
    model_config = ConfigDict(extra="allow")



"""
Module defining models for representing configuration settings for
UI tabs.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from functools import cached_property

from pydantic import BaseModel

from ultimate_rvc.typing_extra import SegmentSize, SeparationModel
from ultimate_rvc.web.config.component import (
    AnyComponentConfig,
    AudioConfig,
    CheckboxConfig,
    ComponentConfig,
    DropdownConfig,
    RadioConfig,
    SliderConfig,
)
from ultimate_rvc.web.config.tab import (
    SongGenerationConfig,
    SpeechGenerationConfig,
    TrainingConfig,
)

if TYPE_CHECKING:
    import gradio as gr


class SongIntermediateAudioConfig(BaseModel):
    """
    Configuration settings for intermediate audio components in the
    one-click song generation tab.

    Attributes
    ----------
    song : AudioConfig
        Configuration settings for the input song audio component.
    vocals : AudioConfig
        Configuration settings for the vocals audio component.
    instrumentals : AudioConfig
        Configuration settings for the instrumentals audio component.
    main_vocals : AudioConfig
        Configuration settings for the main vocals audio component.
    backup_vocals : AudioConfig
        Configuration settings for the backup vocals audio component.
    main_vocals_dereverbed : AudioConfig
        Configuration settings for the main vocals de-reverbed audio
        component.
    main_vocals_reverb : AudioConfig
        Configuration settings for the main vocals reverb audio
        component.
    converted_vocals : AudioConfig
        Configuration settings for the converted vocals audio
        component.
    postprocessed_vocals : AudioConfig
        Configuration settings for the postprocessed vocals audio
        component.
    instrumentals_shifted : AudioConfig
        Configuration settings for the shifted instrumentals audio
        component.
    backup_vocals_shifted : AudioConfig
        Configuration settings for the shifted backup vocals audio
        component.
    all : list[gr.Audio]
        List of instances of all intermediate audio components.

    """

    song: AudioConfig = AudioConfig.intermediate(label="Song")
    vocals: AudioConfig = AudioConfig.intermediate(label="Vocals")
    instrumentals: AudioConfig = AudioConfig.intermediate(
        label="Instrumentals",
    )
    main_vocals: AudioConfig = AudioConfig.intermediate(
        label="Main vocals",
    )
    backup_vocals: AudioConfig = AudioConfig.intermediate(
        label="Backup vocals",
    )
    main_vocals_dereverbed: AudioConfig = AudioConfig.intermediate(
        label="De-reverbed main vocals",
    )
    main_vocals_reverb: AudioConfig = AudioConfig.intermediate(
        label="Main vocals with reverb",
    )
    converted_vocals: AudioConfig = AudioConfig.intermediate(
        label="Converted vocals",
    )
    postprocessed_vocals: AudioConfig = AudioConfig.intermediate(
        label="Postprocessed vocals",
    )
    instrumentals_shifted: AudioConfig = AudioConfig.intermediate(
        label="Pitch-shifted instrumentals",
    )
    backup_vocals_shifted: AudioConfig = AudioConfig.intermediate(
        label="Pitch-shifted backup vocals",
    )

    @property
    def all(self) -> list[gr.Audio]:
        """
        Retrieve instances of all intermediate audio components
        in the one-click song generation tab.

        Returns
        -------
        list[gr.Audio]
            List of instances of all intermediate audio components in
            the one-click song generation tab.

        """
        # NOTE we are using self.__annotations__ to get the fields in
        # the order they are defined in the class
        return [getattr(self, field).instance for field in self.__annotations__]


class OneClickSongGenerationConfig(SongGenerationConfig):
    """
    Configuration settings for the one-click song generation tab.

    Attributes
    ----------
    n_octaves : SliderConfig
        Configuration settings for an octave pitch shift slider
        component.
    n_semitones : SliderConfig
        Configuration settings for a semitone pitch shift slider
        component.
    show_intermediate_audio : CheckboxConfig
        Configuration settings for a show intermediate audio checkbox
        component.
    intermediate_audio : SongIntermediateAudioConfig
        Configuration settings for intermediate audio components.

    See Also
    --------
    SongGenerationConfig
        Parent model defining common component configuration settings
        for song generation tabs.

    """

    n_octaves: SliderConfig = SliderConfig.octave_shift(
        label="Vocal pitch shift",
        info=(
            "The number of octaves to shift the pitch of the converted vocals by. Use 1"
            " for male-to-female and -1 for vice-versa."
        ),
    )

    n_semitones: SliderConfig = SliderConfig.semitone_shift(
        label="Overall pitch shift",
        info=(
            "The number of semi-tones to shift the pitch of the converted vocals,"
            " instrumentals and backup vocals by."
        ),
    )
    show_intermediate_audio: CheckboxConfig = CheckboxConfig(
        label="Show intermediate audio",
        info="Show intermediate audio tracks produced during song cover generation.",
        value=False,
        exclude_value=True,
    )
    intermediate_audio: SongIntermediateAudioConfig = SongIntermediateAudioConfig()


class SongInputAudioConfig(BaseModel):
    """
    Configuration settings for input audio components in the multi-step
    song generation tab.

    Attributes
    ----------
    audio : AudioConfig
        Configuration settings for the input audio component.
    vocals : AudioConfig
        Configuration settings for the vocals audio component.
    converted_vocals : AudioConfig
        Configuration settings for the converted vocals audio
        component.
    instrumentals : AudioConfig
        Configuration settings for the instrumentals audio
        component.
    backup_vocals : AudioConfig
        Configuration settings for the backup vocals audio
        component.
    main_vocals : AudioConfig
        Configuration settings for the main vocals audio
        component.
    shifted_instrumentals : AudioConfig
        Configuration settings for the shifted instrumentals audio
        component.
    shifted_backup_vocals : AudioConfig
        Configuration settings for the shifted backup vocals audio
        component.
    all : list[AudioConfig]
        List of configuration settings for all input audio
        components in the multi-step song generation tab.

    """

    audio: AudioConfig = AudioConfig.input(label="Audio")
    vocals: AudioConfig = AudioConfig.input(label="Vocals")
    converted_vocals: AudioConfig = AudioConfig.input(label="Vocals")
    instrumentals: AudioConfig = AudioConfig.input(label="Instrumentals")
    backup_vocals: AudioConfig = AudioConfig.input(label="Backup vocals")
    main_vocals: AudioConfig = AudioConfig.input(label="Main vocals")
    shifted_instrumentals: AudioConfig = AudioConfig.input(label="Instrumentals")
    shifted_backup_vocals: AudioConfig = AudioConfig.input(label="Backup vocals")

    @property
    def all(self) -> list[AudioConfig]:
        """
        Retrieve configuration settings for all input audio components
        in the multi-step song generation tab.

        Returns
        -------
        list[AudioConfig]
            List of configuration settings for all input audio
            components in the multi-step song generation tab.

        """
        return [getattr(self, field) for field in self.__annotations__]


class SongDirsConfig(BaseModel):
    """
    Configuration settings for song directory components in the
    multi-step song generation tab.

    Attributes
    ----------
    separate_audio : DropdownConfig
        Configuration settings for the song directory component
        for separating audio.
    convert_vocals : DropdownConfig
        Configuration settings for the song directory component
        for converting vocals.
    postprocess_vocals : DropdownConfig
        Configuration settings for the song directory component
        for postprocessing vocals.
    pitch_shift_background : DropdownConfig
        Configuration settings for the song directory component
        for pitch-shifting background audio.
    mix : DropdownConfig
        Configuration settings for the song directory component
        for mixing audio.
    all : list[gr.Dropdown]
        List of instances of all song directory components in the
        multi-step song generation tab.

    """

    separate_audio: DropdownConfig = DropdownConfig.song_dir()
    convert_vocals: DropdownConfig = DropdownConfig.song_dir()
    postprocess_vocals: DropdownConfig = DropdownConfig.song_dir()
    pitch_shift_background: DropdownConfig = DropdownConfig.song_dir()
    mix: DropdownConfig = DropdownConfig.song_dir()

    @property
    def all(self) -> list[gr.Dropdown]:
        """
        Retrieve instances of all song directory components in the
        multi-step song generation tab.

        Returns
        -------
        list[gr.Dropdown]
            List of instances of all song directory components in
            the multi-step song generation tab.

        """
        return [getattr(self, field).instance for field in self.__annotations__]


class MultiStepSongGenerationConfig(SongGenerationConfig):
    """
    Configuration settings for multi-step song generation tab.

    Attributes
    ----------
    separation_model : DropdownConfig
        Configuration settings for a separation model dropdown
        component.
    segment_size : RadioConfig
        Configuration settings for a segment size radio component.
    n_octaves : SliderConfig
        Configuration settings for an octave pitch shift slider
        component.
    n_semitones : SliderConfig
        Configuration settings for a semitone pitch shift slider
        component.
    n_semitones_instrumentals : SliderConfig
        Configuration settings for an instrumentals pitch shift slider
        component.
    n_semitones_backup_vocals : SliderConfig
        Configuration settings for a backup vocals pitch shift slider
        component.
    input_audio : SongInputAudioConfig
        Configuration settings for input audio components.
    song_dirs : SongDirsConfig
        Configuration settings for song directory components.

    See Also
    --------
    SongGenerationConfig
        Parent model defining common component configuration settings
        for song generation tabs.

    """

    separation_model: DropdownConfig = DropdownConfig(
        label="Separation model",
        info="The model to use for audio separation.",
        value=SeparationModel.UVR_MDX_NET_VOC_FT,
        choices=list(SeparationModel),
    )
    segment_size: RadioConfig = RadioConfig(
        label="Segment size",
        info=(
            "The size of the segments into which the audio is split. Using a larger"
            " size consumes more resources, but may give better results."
        ),
        value=SegmentSize.SEG_512,
        choices=list(SegmentSize),
    )
    n_octaves: SliderConfig = SliderConfig.octave_shift(
        label="Pitch shift (octaves)",
        info=(
            "The number of octaves to pitch-shift the converted voice by. Use 1 for"
            " male-to-female and -1 for vice-versa."
        ),
    )
    n_semitones: SliderConfig = SliderConfig.semitone_shift(
        label="Pitch shift (semi-tones)",
        info=(
            "The number of semi-tones to pitch-shift the converted vocals by. Altering"
            " this slightly reduces sound quality."
        ),
    )
    n_semitones_instrumentals: SliderConfig = SliderConfig.semitone_shift(
        label="Instrumental pitch shift",
        info="The number of semi-tones to pitch-shift the instrumentals by.",
    )
    n_semitones_backup_vocals: SliderConfig = SliderConfig.semitone_shift(
        label="Backup vocal pitch shift",
        info="The number of semi-tones to pitch-shift the backup vocals by.",
    )
    input_audio: SongInputAudioConfig = SongInputAudioConfig()
    song_dirs: SongDirsConfig = SongDirsConfig()


class SpeechIntermediateAudioConfig(BaseModel):
    """
    Configuration settings for intermediate audio components in the
    one-click speech generation tab.

    Attributes
    ----------
    speech : AudioConfig
        Configuration settings for the input speech audio component.
    converted_speech : AudioConfig
        Configuration settings for the converted speech audio component.
    all : list[gr.Audio]
        List of instances of all intermediate audio components in the
        speech generation tab.

    """

    speech: AudioConfig = AudioConfig.intermediate(label="Speech")
    converted_speech: AudioConfig = AudioConfig.intermediate(label="Converted speech")

    @property
    def all(self) -> list[gr.Audio]:
        """
        Retrieve instances of all intermediate audio components in the
        speech generation tab.

        Returns
        -------
        list[gr.Audio]
            List of instances of all intermediate audio components in
            the speech generation tab.

        """
        return [getattr(self, field).instance for field in self.__annotations__]


class OneClickSpeechGenerationConfig(SpeechGenerationConfig):
    """
    Configuration settings for one-click speech generation tab.

    Attributes
    ----------
    intermediate_audio : SpeechIntermediateAudioConfig
        Configuration settings for intermediate audio components.
    show_intermediate_audio : CheckboxConfig
        Configuration settings for a show intermediate audio checkbox
        component.

    See Also
    --------
    SpeechGenerationConfig
        Parent model defining common component configuration settings
        for speech generation tabs.

    """

    intermediate_audio: SpeechIntermediateAudioConfig = SpeechIntermediateAudioConfig()

    show_intermediate_audio: CheckboxConfig = CheckboxConfig(
        label="Show intermediate audio",
        info="Show intermediate audio tracks produced during speech generation.",
        value=False,
        exclude_value=True,
    )


class SpeechInputAudioConfig(BaseModel):
    """
    Configuration settings for input audio components in the multi-step
    speech generation tab.

    Attributes
    ----------
    speech : AudioConfig
        Configuration settings for the input speech audio component.
    converted_speech : AudioConfig
        Configuration settings for the converted speech audio component.

    all : list[AudioConfig]
        List of configuration settings for all input audio components in
        the multi-step speech generation tab.

    """

    speech: AudioConfig = AudioConfig.input("Speech")
    converted_speech: AudioConfig = AudioConfig.input("Converted speech")

    @property
    def all(self) -> list[AudioConfig]:
        """
        Retrieve configuration settings for all input audio components
        in the multi-step speech generation tab.

        Returns
        -------
        list[AudioConfig]
            List of configuration settings for all input audio
            components in the multi-step speech generation tab.

        """
        return [getattr(self, field) for field in self.__annotations__]


class MultiStepSpeechGenerationConfig(SpeechGenerationConfig):
    """
    Configuration settings for the multi-step speech generation tab.

    Attributes
    ----------
    input_audio : SpeechInputAudioConfig
        Configuration settings for input audio components.

    See Also
    --------
    SpeechGenerationConfig
        Parent model defining common component configuration settings
        for speech generation tabs.

    """

    input_audio: SpeechInputAudioConfig = SpeechInputAudioConfig()


class MultiStepTrainingConfig(TrainingConfig):
    """Configuration settings for multi-step training tab."""


class ModelManagementConfig(BaseModel):
    """

    Configuration settings for model management tab.

    Attributes
    ----------
    voices : DropdownConfig
        Configuration settings for delete voice models dropdown
        component.
    embedders : DropdownConfig
        Configuration settings for delete embedder models dropdown
        component.
    pretraineds : DropdownConfig
        Configuration settings for delete pretrained models dropdown
        component.
    traineds : DropdownConfig
        Configuration settings for delete training models dropdown
        component.
    dummy_checkbox : CheckboxConfig
        Configuration settings for a dummy checkbox component.

    """

    voices: DropdownConfig = DropdownConfig.multi_delete(
        label="Voice models",
        info="Select one or more voice models to delete.",
    )
    embedders: DropdownConfig = DropdownConfig.multi_delete(
        label="Custom embedder models",
        info="Select one or more embedder models to delete.",
    )
    pretraineds: DropdownConfig = DropdownConfig.multi_delete(
        label="Custom pretrained models",
        info="Select one or more pretrained models to delete.",
    )
    traineds: DropdownConfig = DropdownConfig.multi_delete(
        label="Training models",
        info="Select one or more training models to delete.",
    )

    dummy_checkbox: CheckboxConfig = CheckboxConfig(
        value=False,
        visible=False,
        exclude_value=True,
    )


class AudioManagementConfig(BaseModel):
    """
    Configuration settings for audio management tab.

    Attributes
    ----------
    intermediate : DropdownConfig
        Configuration settings for delete intermediate audio files
        dropdown component
    speech : DropdownConfig
        Configuration settings for delete speech audio files dropdown
        component.
    output : DropdownConfig
        Configuration settings for delete output audio files dropdown
        component.
    dataset : DropdownConfig
        Configuration settings for delete dataset audio files dropdown
        component.
    dummy_checkbox : CheckboxConfig
        Configuration settings for a dummy checkbox component.

    """

    intermediate: DropdownConfig = DropdownConfig.multi_delete(
        label="Song directories",
        info=(
            "Select one or more song directories containing intermediate audio files to"
            " delete."
        ),
    )
    speech: DropdownConfig = DropdownConfig.multi_delete(
        label="Speech audio files",
        info="Select one or more speech audio files to delete.",
    )
    output: DropdownConfig = DropdownConfig.multi_delete(
        label="Output audio files",
        info="Select one or more output audio files to delete.",
    )
    dataset: DropdownConfig = DropdownConfig.multi_delete(
        label="Dataset audio files",
        info="Select one or more datasets containing audio files to delete.",
    )

    dummy_checkbox: CheckboxConfig = CheckboxConfig(
        value=False,
        visible=False,
        exclude_value=True,
    )


class SettingsManagementConfig(BaseModel):
    """
    Configuration settings for settings management tab.

    Attributes
    ----------
    dummy_checkbox : CheckboxConfig
        Configuration settings for a dummy checkbox component.

    """

    load_config_name: DropdownConfig = DropdownConfig(
        label="Configuration name",
        info="The name of a configuration to load UI settings from",
        value=None,
        render=False,
        exclude_value=True,
    )
    delete_config_names: DropdownConfig = DropdownConfig.multi_delete(
        label="Configuration names",
        info="Select the name of one or more configurations to delete",
    )
    dummy_checkbox: CheckboxConfig = CheckboxConfig(
        value=False,
        visible=False,
        exclude_value=True,
    )


class TotalSongGenerationConfig(BaseModel):
    """
    All configuration settings for song generation tabs.

    Attributes
    ----------
    one_click : OneClickSongGenerationConfig
        Configuration settings for the one-click song generation tab.
    multi_step : MultiStepSongGenerationConfig
        Configuration settings for the multi-step song generation tab.

    """

    one_click: OneClickSongGenerationConfig = OneClickSongGenerationConfig()
    multi_step: MultiStepSongGenerationConfig = MultiStepSongGenerationConfig()


class TotalSpeechGenerationConfig(BaseModel):
    """
    All configuration settings for speech generation tabs.

    Attributes
    ----------
    one_click : OneClickSpeechGenerationConfig
        Configuration settings for the one-click speech generation tab.
    multi_step : MultiStepSpeechGenerationConfig
        Configuration settings for the multi-step speech generation tab.

    """

    one_click: OneClickSpeechGenerationConfig = OneClickSpeechGenerationConfig()
    multi_step: MultiStepSpeechGenerationConfig = MultiStepSpeechGenerationConfig()


class TotalTrainingConfig(BaseModel):
    """
    All configuration settings for training tabs.

    Attributes
    ----------
    training : TrainingConfig
        Configuration settings for the multi-step training tab.

    """

    multi_step: MultiStepTrainingConfig = MultiStepTrainingConfig()


class TotalManagementConfig(BaseModel):
    """
    All configuration settings for management tabs.

    Attributes
    ----------
    model : ModelManagementConfig
        Configuration settings for the model management tab.
    audio : AudioManagementConfig
        Configuration settings for the audio management tab.
    settings : SettingsManagementConfig
        Configuration settings for the settings management tab.

    """

    model: ModelManagementConfig = ModelManagementConfig()
    audio: AudioManagementConfig = AudioManagementConfig()
    settings: SettingsManagementConfig = SettingsManagementConfig()


class TotalConfig(BaseModel):
    """
    All configuration settings for the Ultimate RVC app.

    Attributes
    ----------
    song : TotalSongGenerationConfig
        Configuration settings for song generation tabs.
    speech : TotalSpeechGenerationConfig
        Configuration settings for speech generation tabs.
    training : TotalTrainingConfig
        Configuration settings for training tabs.
    management : TotalManagementConfig
        Configuration settings for management tabs.

    """

    song: TotalSongGenerationConfig = TotalSongGenerationConfig()
    speech: TotalSpeechGenerationConfig = TotalSpeechGenerationConfig()
    training: TotalTrainingConfig = TotalTrainingConfig()
    management: TotalManagementConfig = TotalManagementConfig()

    @cached_property
    def all(self) -> list[AnyComponentConfig]:
        """
        Recursively collect those component configuration models nested
        within the current model instance, which have values that are
        not excluded.

        Returns
        -------
        list[AnyComponentConfig]
            A list of component configuration models found within the
            current model instance, which have values that are not
            excluded.

        """

        def _collect(model: BaseModel) -> list[AnyComponentConfig]:
            component_configs: list[Any] = []
            for _, value in model:
                if isinstance(value, ComponentConfig):
                    if not value.exclude_value:
                        component_configs.append(value)
                elif isinstance(value, BaseModel):
                    component_configs.extend(_collect(value))
            return component_configs

        return _collect(self)



"""
Web application for the Ultimate RVC project.

Each tab of the application is defined in its own module in the
`web/tabs` directory. Components that are accessed across multiple
tabs are passed as arguments to the render functions in the respective
modules.
"""

from __future__ import annotations

from typing import Annotated

import os

import gradio as gr

import typer

from ultimate_rvc.common import AUDIO_DIR, MODELS_DIR, TEMP_DIR
from ultimate_rvc.core.generate.song_cover import get_named_song_dirs
from ultimate_rvc.core.generate.speech import get_edge_tts_voice_names
from ultimate_rvc.core.manage.audio import (
    get_audio_datasets,
    get_named_audio_datasets,
    get_saved_output_audio,
    get_saved_speech_audio,
)
from ultimate_rvc.core.manage.config import get_config_names, load_config
from ultimate_rvc.core.manage.models import (
    get_custom_embedder_model_names,
    get_custom_pretrained_model_names,
    get_training_model_names,
    get_voice_model_names,
)
from ultimate_rvc.web.common import initialize_dropdowns
from ultimate_rvc.web.config.main import TotalConfig
from ultimate_rvc.web.tabs.generate.song_cover.multi_step_generation import (
    render as render_song_cover_multi_step_tab,
)
from ultimate_rvc.web.tabs.generate.song_cover.one_click_generation import (
    render as render_song_cover_one_click_tab,
)
from ultimate_rvc.web.tabs.generate.speech.multi_step_generation import (
    render as render_speech_multi_step_tab,
)
from ultimate_rvc.web.tabs.generate.speech.one_click_generation import (
    render as render_speech_one_click_tab,
)
from ultimate_rvc.web.tabs.manage.audio import render as render_audio_tab
from ultimate_rvc.web.tabs.manage.models import render as render_models_tab
from ultimate_rvc.web.tabs.manage.settings import render as render_settings_tab

config_name = os.environ.get("URVC_CONFIG")
cookiefile = os.environ.get("YT_COOKIEFILE")
total_config = load_config(config_name, TotalConfig) if config_name else TotalConfig()


def render_app() -> gr.Blocks:
    """
    Render the Ultimate RVC web application.

    Returns
    -------
    gr.Blocks
        The rendered web application.

    """
    css = """
    h1 { text-align: center; margin-top: 20px; margin-bottom: 20px; }

    #generate-tab-button { font-weight: bold !important;}
    #manage-tab-button { font-weight: bold !important;}
    #audio-tab-button { font-weight: bold !important;}
    #settings-tab-button { font-weight: bold !important;}
    """
    cache_delete_frequency = 86400  # every 24 hours check for files to delete
    cache_delete_cutoff = 86400  # and delete files older than 24 hours

    with gr.Blocks(
        title="Ultimate RVC",
        theme="JackismyShephard/ultimate-rvc-theme",
        css=css,
        delete_cache=(cache_delete_frequency, cache_delete_cutoff),
    ) as app:
        gr.HTML("<h1>Ultimate RVC 💙</h1>")
        for component_config in [
            total_config.song.one_click.voice_model,
            total_config.song.one_click.cached_song,
            total_config.song.one_click.custom_embedder_model,
            total_config.song.multi_step.voice_model,
            total_config.song.multi_step.cached_song,
            total_config.song.multi_step.custom_embedder_model,
            total_config.song.multi_step.song_dirs.separate_audio,
            total_config.song.multi_step.song_dirs.convert_vocals,
            total_config.song.multi_step.song_dirs.postprocess_vocals,
            total_config.song.multi_step.song_dirs.pitch_shift_background,
            total_config.song.multi_step.song_dirs.mix,
            total_config.speech.one_click.edge_tts_voice,
            total_config.speech.one_click.voice_model,
            total_config.speech.one_click.custom_embedder_model,
            total_config.speech.multi_step.edge_tts_voice,
            total_config.speech.multi_step.voice_model,
            total_config.speech.multi_step.custom_embedder_model,
            total_config.training.multi_step.dataset,
            total_config.training.multi_step.preprocess_model,
            total_config.training.multi_step.extract_model,
            total_config.training.multi_step.train_model,
            total_config.training.multi_step.custom_embedder_model,
            total_config.training.multi_step.custom_pretrained_model,
            total_config.management.audio.intermediate,
            total_config.management.audio.speech,
            total_config.management.audio.output,
            total_config.management.audio.dataset,
            total_config.management.model.voices,
            total_config.management.model.embedders,
            total_config.management.model.pretraineds,
            total_config.management.model.traineds,
            total_config.management.settings.load_config_name,
            total_config.management.settings.delete_config_names,
        ]:
            component_config.instantiate()
        # main tab
        with gr.Tab("Generate", elem_id="generate-tab"):
            with gr.Tab("Song covers"):
                render_song_cover_one_click_tab(total_config, cookiefile)
                render_song_cover_multi_step_tab(total_config, cookiefile)
            with gr.Tab("Speech"):
                render_speech_one_click_tab(total_config)
                render_speech_multi_step_tab(total_config)
        with gr.Tab("Models", elem_id="manage-tab"):
            render_models_tab(total_config)
        with gr.Tab("Audio", elem_id="audio-tab"):
            render_audio_tab(total_config)
        with gr.Tab("Settings", elem_id="settings-tab"):
            render_settings_tab(total_config)

        app.load(
            _init_dropdowns,
            outputs=[
                total_config.speech.one_click.edge_tts_voice.instance,
                total_config.speech.multi_step.edge_tts_voice.instance,
                total_config.song.one_click.voice_model.instance,
                total_config.song.multi_step.voice_model.instance,
                total_config.speech.one_click.voice_model.instance,
                total_config.speech.multi_step.voice_model.instance,
                total_config.management.model.voices.instance,
                total_config.song.one_click.custom_embedder_model.instance,
                total_config.song.multi_step.custom_embedder_model.instance,
                total_config.speech.one_click.custom_embedder_model.instance,
                total_config.speech.multi_step.custom_embedder_model.instance,
                total_config.training.multi_step.custom_embedder_model.instance,
                total_config.management.model.embedders.instance,
                total_config.training.multi_step.custom_pretrained_model.instance,
                total_config.management.model.pretraineds.instance,
                total_config.training.multi_step.extract_model.instance,
                total_config.training.multi_step.train_model.instance,
                total_config.training.multi_step.preprocess_model.instance,
                total_config.management.model.traineds.instance,
                total_config.song.one_click.cached_song.instance,
                total_config.song.multi_step.cached_song.instance,
                total_config.song.multi_step.song_dirs.separate_audio.instance,
                total_config.song.multi_step.song_dirs.convert_vocals.instance,
                total_config.song.multi_step.song_dirs.postprocess_vocals.instance,
                total_config.song.multi_step.song_dirs.pitch_shift_background.instance,
                total_config.song.multi_step.song_dirs.mix.instance,
                total_config.management.audio.intermediate.instance,
                total_config.training.multi_step.dataset.instance,
                total_config.management.audio.speech.instance,
                total_config.management.audio.output.instance,
                total_config.management.audio.dataset.instance,
                total_config.management.settings.load_config_name.instance,
                total_config.management.settings.delete_config_names.instance,
            ],
            show_progress="hidden",
        )
    return app


def _init_dropdowns() -> list[gr.Dropdown]:
    """
    Initialize the Ultimate RVC web application by updating the choices
    and default values of non-static dropdown components.

    Returns
    -------
    tuple[gr.Dropdown, ...]
        A tuple of gr.Dropdown components with updated choices and
        default values.

    """
    # Initialize model dropdowns
    edge_tts_models = initialize_dropdowns(
        get_edge_tts_voice_names,
        2,
        "en-US-ChristopherNeural",
        range(2),
    )
    voice_models = initialize_dropdowns(
        get_voice_model_names,
        5,
        value_indices=range(4),
    )
    custom_embedder_models = initialize_dropdowns(
        get_custom_embedder_model_names,
        6,
        value_indices=range(5),
    )
    custom_pretrained_models = initialize_dropdowns(
        get_custom_pretrained_model_names,
        2,
        value_indices=range(1),
    )
    training_models = initialize_dropdowns(
        get_training_model_names,
        4,
        value_indices=range(2),
    )
    song_dirs = initialize_dropdowns(
        get_named_song_dirs,
        8,
        value_indices=range(7),
    )
    dataset = gr.Dropdown(get_audio_datasets())
    speech_delete = gr.Dropdown(get_saved_speech_audio())
    output_delete = gr.Dropdown(get_saved_output_audio())
    dataset_delete = gr.Dropdown(get_named_audio_datasets())
    configs = initialize_dropdowns(get_config_names, 2, value_indices=range(1))
    return [
        *edge_tts_models,
        *voice_models,
        *custom_embedder_models,
        *custom_pretrained_models,
        *training_models,
        *song_dirs,
        dataset,
        speech_delete,
        output_delete,
        dataset_delete,
        *configs,
    ]


app = render_app()
app_wrapper = typer.Typer()


@app_wrapper.command()
def start_app(
    share: Annotated[
        bool,
        typer.Option("--share", "-s", help="Enable sharing"),
    ] = False,
    listen: Annotated[
        bool,
        typer.Option(
            "--listen",
            "-l",
            help="Make the web application reachable from your local network.",
        ),
    ] = False,
    listen_host: Annotated[
        str | None,
        typer.Option(
            "--listen-host",
            "-h",
            help="The hostname that the server will use.",
        ),
    ] = None,
    listen_port: Annotated[
        int | None,
        typer.Option(
            "--listen-port",
            "-p",
            help="The listening port that the server will use.",
        ),
    ] = None,
    ssr_mode: Annotated[
        bool,
        typer.Option(
            "--ssr-mode",
            help="Enable server-side rendering mode.",
        ),
    ] = False,
) -> None:
    """Run the Ultimate RVC web application."""
    os.environ["GRADIO_TEMP_DIR"] = str(TEMP_DIR)
    gr.set_static_paths([MODELS_DIR, AUDIO_DIR])
    app.queue()
    app.launch(
        share=share,
        server_name=(None if not listen else (listen_host or "0.0.0.0")),  # noqa: S104
        server_port=listen_port,
        ssr_mode=ssr_mode,
    )


if __name__ == "__main__":
    app_wrapper()



from typing import Literal

from os import PathLike

import numpy as np
from numpy.typing import NDArray

type DEFAULT_NDARRAY = NDArray[np.float64 | np.float32 | np.int32 | np.int16]

def read(
    file: int | str | PathLike[str] | PathLike[bytes],
    frames: int = -1,
    start: int = 0,
    stop: int | None = None,
    dtype: Literal["float64", "float32", "int32", "int16"] = "float64",
    always_2d: bool = False,
    fill_value: float | None = None,
    out: DEFAULT_NDARRAY | None = None,
    samplerate: int | None = None,
    channels: int | None = None,
    format: str | None = None,  # noqa: A002
    subtype: str | None = None,
    endian: Literal["FILE", "LITTLE", "BIG", "CPU"] | None = None,
    closefd: bool | None = True,
) -> tuple[DEFAULT_NDARRAY, int]: ...
def write(
    file: int | str | PathLike[str] | PathLike[bytes],
    data: DEFAULT_NDARRAY,
    samplerate: int,
    subtype: str | None = None,
    endian: Literal["FILE", "LITTLE", "BIG", "CPU"] | None = None,
    format: str | None = None,  # noqa: A002
    closefd: bool | None = True,
    compression_level: float | None = None,
    bitrate_mode: Literal["CONSTANT", "AVERAGE", "VARIABLE"] | None = None,
) -> None: ...



import logging
import os
import re
import sys
import unicodedata
import warnings

import soxr

import wget

import numpy as np

from torch import nn
from transformers import HubertModel

import librosa
import soundfile as sf

from ultimate_rvc.common import RVC_MODELS_DIR

# Remove this to see warnings about transformers models
warnings.filterwarnings("ignore")

logging.getLogger("fairseq").setLevel(logging.ERROR)
logging.getLogger("faiss.loader").setLevel(logging.ERROR)
logging.getLogger("transformers").setLevel(logging.ERROR)
logging.getLogger("torch").setLevel(logging.ERROR)

now_dir = os.getcwd()
sys.path.append(now_dir)

base_path = os.path.join(str(RVC_MODELS_DIR), "formant", "stftpitchshift")
stft = base_path + ".exe" if sys.platform == "win32" else base_path


class HubertModelWithFinalProj(HubertModel):
    def __init__(self, config):
        super().__init__(config)
        self.final_proj = nn.Linear(config.hidden_size, config.classifier_proj_size)


def load_audio(file, sample_rate):
    try:
        file = file.strip(" ").strip('"').strip("\n").strip('"').strip(" ")
        audio, sr = sf.read(file)
        if len(audio.shape) > 1:
            audio = librosa.to_mono(audio.T)
        if sr != sample_rate:
            audio = librosa.resample(
                audio,
                orig_sr=sr,
                target_sr=sample_rate,
                res_type="soxr_vhq",
            )
    except Exception as error:
        raise RuntimeError(f"An error occurred loading the audio: {error}")

    return audio.flatten()


def load_audio_infer(
    file,
    sample_rate,
    **kwargs,
):
    formant_shifting = kwargs.get("formant_shifting", False)
    try:
        file = file.strip(" ").strip('"').strip("\n").strip('"').strip(" ")
        if not os.path.isfile(file):
            raise FileNotFoundError(f"File not found: {file}")
        audio, sr = sf.read(file)
        if len(audio.shape) > 1:
            audio = librosa.to_mono(audio.T)
        if sr != sample_rate:
            audio = librosa.resample(
                audio,
                orig_sr=sr,
                target_sr=sample_rate,
                res_type="soxr_vhq",
            )
        if formant_shifting:
            formant_qfrency = kwargs.get("formant_qfrency", 0.8)
            formant_timbre = kwargs.get("formant_timbre", 0.8)

            from stftpitchshift import StftPitchShift

            pitchshifter = StftPitchShift(1024, 32, sample_rate)
            audio = pitchshifter.shiftpitch(
                audio,
                factors=1,
                quefrency=formant_qfrency * 1e-3,
                distortion=formant_timbre,
            )
    except Exception as error:
        raise RuntimeError(f"An error occurred loading the audio: {error}")
    return np.array(audio).flatten()


def format_title(title):
    formatted_title = (
        unicodedata.normalize("NFKD", title).encode("ascii", "ignore").decode("utf-8")
    )
    formatted_title = re.sub(r"[\u2500-\u257F]+", "", formatted_title)
    formatted_title = re.sub(r"[^\w\s.-]", "", formatted_title)
    formatted_title = re.sub(r"\s+", "_", formatted_title)
    return formatted_title


def load_embedding(embedder_model, custom_embedder=None):
    embedder_root = os.path.join(str(RVC_MODELS_DIR), "embedders")
    embedding_list = {
        "contentvec": os.path.join(embedder_root, "contentvec"),
        "chinese-hubert-base": os.path.join(embedder_root, "chinese_hubert_base"),
        "japanese-hubert-base": os.path.join(embedder_root, "japanese_hubert_base"),
        "korean-hubert-base": os.path.join(embedder_root, "korean_hubert_base"),
    }

    online_embedders = {
        "contentvec": "https://huggingface.co/IAHispano/Applio/resolve/main/Resources/embedders/contentvec/pytorch_model.bin",
        "chinese-hubert-base": "https://huggingface.co/IAHispano/Applio/resolve/main/Resources/embedders/chinese_hubert_base/pytorch_model.bin",
        "japanese-hubert-base": "https://huggingface.co/IAHispano/Applio/resolve/main/Resources/embedders/japanese_hubert_base/pytorch_model.bin",
        "korean-hubert-base": "https://huggingface.co/IAHispano/Applio/resolve/main/Resources/embedders/korean_hubert_base/pytorch_model.bin",
    }

    config_files = {
        "contentvec": "https://huggingface.co/IAHispano/Applio/resolve/main/Resources/embedders/contentvec/config.json",
        "chinese-hubert-base": "https://huggingface.co/IAHispano/Applio/resolve/main/Resources/embedders/chinese_hubert_base/config.json",
        "japanese-hubert-base": "https://huggingface.co/IAHispano/Applio/resolve/main/Resources/embedders/japanese_hubert_base/config.json",
        "korean-hubert-base": "https://huggingface.co/IAHispano/Applio/resolve/main/Resources/embedders/korean_hubert_base/config.json",
    }

    if embedder_model == "custom":
        if os.path.exists(custom_embedder):
            model_path = custom_embedder
        else:
            print(f"Custom embedder not found: {custom_embedder}, using contentvec")
            model_path = embedding_list["contentvec"]
    else:
        model_path = embedding_list[embedder_model]
        bin_file = os.path.join(model_path, "pytorch_model.bin")
        json_file = os.path.join(model_path, "config.json")
        os.makedirs(model_path, exist_ok=True)
        if not os.path.exists(bin_file):
            url = online_embedders[embedder_model]
            print(f"Downloading {url} to {model_path}...")
            wget.download(url, out=bin_file)
        if not os.path.exists(json_file):
            url = config_files[embedder_model]
            print(f"Downloading {url} to {model_path}...")
            wget.download(url, out=json_file)

    models = HubertModelWithFinalProj.from_pretrained(model_path)
    return models



"""
Common definitions for modules in the Ultimate RVC project that
facilitate training voice models.
"""

from __future__ import annotations

from typing import Literal

from ultimate_rvc.core.exceptions import (
    Entity,
    GPUNotFoundError,
    NotProvidedError,
    UIMessage,
)
from ultimate_rvc.typing_extra import DeviceType


def get_gpu_info() -> list[tuple[str, int]]:
    """
    Retrieve information on locally available GPUs.

    Returns
    -------
    list[tuple[str, int]]
        A list of tuples containing the name and index of each locally
        available GPU.

    """
    # NOTE lazy importing does not work with torch so we import it here
    #  manually
    import torch  # noqa: PLC0415

    ngpu = torch.cuda.device_count()
    gpu_infos: list[tuple[str, int]] = []
    if torch.cuda.is_available() or ngpu != 0:
        for i in range(ngpu):
            gpu_name = torch.cuda.get_device_name(i)
            mem = int(
                torch.cuda.get_device_properties(i).total_memory / 1024 / 1024 / 1024  # type: ignore[ReportUnknownMembershipType]
                + 0.4,
            )
            gpu_infos.append((f"{gpu_name} ({mem} GB)", i))
    return gpu_infos


def validate_devices(
    device_type: DeviceType,
    device_ids: set[int] | None = None,
) -> tuple[Literal["cuda", "cpu"], set[int] | None]:
    """
    Validate the devices identified by the provided device type and
    device IDs.

    If the provided device type is AUTOMATIC, the first available GPU
    will be selected if available. Otherwise CPU will be selected.
    If the device type is GPU, then validation will be performed to
    ensure that at least one device ID is provided and that all device
    IDs point to available GPUs. If the device type is CPU, then no
    validation is performed.

    Parameters
    ----------
    device_type : DeviceType
        The type of devices to validate.
    device_ids : set[int], optional
        The IDs of the devices to validate when device type is GPU.

    Returns
    -------
        device_type : str
            The type of the selected devices.
        device_ids : set[int], optional
            The ids of the selected devices. Only returned when the
            device type is GPU or AUTOMATIC.

    Raises
    ------
    NotProvidedError
        If device type is GPU and no device IDs are provided.
    GPUNotFoundError
        If device type is GPU and a provided device ID does not point
        to an available GPU.


    """
    match device_type:
        case DeviceType.AUTOMATIC:
            gpu_info = get_gpu_info()
            if gpu_info:
                return "cuda", {gpu_info[0][1]}
            return "cpu", None
        case DeviceType.GPU:
            if not device_ids:
                raise NotProvidedError(Entity.GPU_IDS, UIMessage.NO_GPUS)
            validated_devices: list[int] = []
            available_ids = {i for _, i in get_gpu_info()}
            for device_id in device_ids:
                if device_id not in available_ids:
                    raise GPUNotFoundError(device_id)
                validated_devices.append(device_id)
            return "cuda", set(validated_devices)
        case DeviceType.CPU:
            return "cpu", None



from typing import TYPE_CHECKING

import lazy_loader as lazy

import os
from concurrent.futures import ThreadPoolExecutor

from ultimate_rvc.common import (
    EMBEDDER_MODELS_DIR,
    PRETRAINED_MODELS_DIR,
    RVC_MODELS_DIR,
)

if TYPE_CHECKING:
    import requests

    import tqdm

else:
    tqdm = lazy.load("tqdm")
    requests = lazy.load("requests")


url_base = "https://huggingface.co/JackismyShephard/ultimate-rvc/resolve/main/Resources"

pretraineds_hifigan_list = [
    (
        "pretrained_v2/",
        [
            "f0D32k.pth",
            "f0D40k.pth",
            "f0D48k.pth",
            "f0G32k.pth",
            "f0G40k.pth",
            "f0G48k.pth",
        ],
    ),
]
models_list = [("predictors/", ["rmvpe.pt", "fcpe.pt"])]
embedders_list = [
    ("embedders/contentvec/", ["pytorch_model.bin", "config.json"]),
    ("embedders/chinese_hubert_base/", ["pytorch_model.bin", "config.json"]),
    ("embedders/japanese_hubert_base/", ["pytorch_model.bin", "config.json"]),
    ("embedders/korean_hubert_base/", ["pytorch_model.bin", "config.json"]),
]
executables_list = [
    ("", ["ffmpeg.exe", "ffprobe.exe"]),
]

folder_mapping_list = {
    "pretrained_v2/": str(PRETRAINED_MODELS_DIR / "hifi-gan/"),
    "embedders/contentvec/": str(EMBEDDER_MODELS_DIR / "contentvec/"),
    "embedders/chinese_hubert_base/": str(
        EMBEDDER_MODELS_DIR / "chinese_hubert_base/",
    ),
    "embedders/japanese_hubert_base/": str(
        EMBEDDER_MODELS_DIR / "japanese_hubert_base/",
    ),
    "embedders/korean_hubert_base/": str(
        EMBEDDER_MODELS_DIR / "korean_hubert_base/",
    ),
    "predictors/": str(RVC_MODELS_DIR / "predictors/"),
    "formant/": str(RVC_MODELS_DIR / "formant/"),
}


def get_file_size_if_missing(file_list):
    """
    Calculate the total size of files to be downloaded only if they do not exist locally.
    """
    total_size = 0
    for remote_folder, files in file_list:
        local_folder = folder_mapping_list.get(remote_folder, "")
        for file in files:
            destination_path = os.path.join(local_folder, file)
            if not os.path.exists(destination_path):
                url = f"{url_base}/{remote_folder}{file}"
                response = requests.head(url)
                total_size += int(response.headers.get("content-length", 0))
    return total_size


def download_file(url, destination_path, global_bar):
    """
    Download a file from the given URL to the specified destination path,
    updating the global progress bar as data is downloaded.
    """
    dir_name = os.path.dirname(destination_path)
    if dir_name:
        os.makedirs(dir_name, exist_ok=True)
    response = requests.get(url, stream=True)
    block_size = 1024
    with open(destination_path, "wb") as file:
        for data in response.iter_content(block_size):
            file.write(data)
            global_bar.update(len(data))


def download_mapping_files(file_mapping_list, global_bar):
    """
    Download all files in the provided file mapping list using a thread pool executor,
    and update the global progress bar as downloads progress.
    """
    with ThreadPoolExecutor() as executor:
        futures = []
        for remote_folder, file_list in file_mapping_list:
            local_folder = folder_mapping_list.get(remote_folder, "")
            for file in file_list:
                destination_path = os.path.join(local_folder, file)
                if not os.path.exists(destination_path):
                    url = f"{url_base}/{remote_folder}{file}"
                    futures.append(
                        executor.submit(
                            download_file,
                            url,
                            destination_path,
                            global_bar,
                        ),
                    )
        for future in futures:
            future.result()


def split_pretraineds(pretrained_list):
    f0_list = []
    non_f0_list = []
    for folder, files in pretrained_list:
        f0_files = [f for f in files if f.startswith("f0")]
        non_f0_files = [f for f in files if not f.startswith("f0")]
        if f0_files:
            f0_list.append((folder, f0_files))
        if non_f0_files:
            non_f0_list.append((folder, non_f0_files))
    return f0_list, non_f0_list


pretraineds_hifigan_list, _ = split_pretraineds(pretraineds_hifigan_list)


def calculate_total_size(
    pretraineds_hifigan,
    models,
    exe,
):
    """
    Calculate the total size of all files to be downloaded based on selected categories.
    """
    total_size = 0
    if models:
        total_size += get_file_size_if_missing(models_list)
        total_size += get_file_size_if_missing(embedders_list)
    if exe and os.name == "nt":
        total_size += get_file_size_if_missing(executables_list)
    total_size += get_file_size_if_missing(pretraineds_hifigan)
    return total_size


def prequisites_download_pipeline(
    pretraineds_hifigan: bool = True,
    models: bool = True,
    exe: bool = True,
) -> None:
    """
    Manage the download pipeline for different categories of files.
    """
    total_size = calculate_total_size(
        pretraineds_hifigan_list if pretraineds_hifigan else [],
        models,
        exe,
    )

    if total_size > 0:
        with tqdm.tqdm(
            total=total_size,
            unit="iB",
            unit_scale=True,
            desc="Downloading all files",
        ) as global_bar:
            if models:
                download_mapping_files(models_list, global_bar)
                download_mapping_files(embedders_list, global_bar)
            if exe:
                if os.name == "nt":
                    download_mapping_files(executables_list, global_bar)
                else:
                    print("No executables needed")
            if pretraineds_hifigan:
                download_mapping_files(pretraineds_hifigan_list, global_bar)



"""Module which defines functions to manage configuration files."""

from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from pathlib import Path

from ultimate_rvc.common import CONFIG_DIR
from ultimate_rvc.core.common import json_dump, json_load
from ultimate_rvc.core.exceptions import (
    ConfigExistsError,
    ConfigNotFoundError,
    Entity,
    NotProvidedError,
    UIMessage,
)
from ultimate_rvc.core.manage.common import delete_directory, get_items

if TYPE_CHECKING:
    from collections.abc import Sequence

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


def get_config_names() -> list[str]:
    """
    Get the names of all configuration files.

    Returns
    -------
    list[str]
        A list containing the names of all configuration files.


    """
    return get_items(CONFIG_DIR)


def load_config(name: str, config_class: type[T]) -> T:  # noqa: UP047
    """
    Load a configuration from a JSON file into a Pydantic model.

    Parameters
    ----------
    name: str
        The name of the configuration to load.
    config_class: type[T]
        The class of pydantic model to load the configuration into.


    Returns
    -------
    T
        The Pydantic model containing the configuration.

    Raises
    ------
    NotProvidedError
        If no name is provided for the configuration to load.
    ConfigNotFoundError
        If a configuration with the provided name does not exist.

    """
    if not name:
        raise NotProvidedError(entity=Entity.CONFIG_NAME)
    name = name.strip()
    file_path = CONFIG_DIR / f"{name}.json"
    if not file_path.is_file():
        raise ConfigNotFoundError(name=name)
    config_data = json_load(file_path)
    return config_class.model_validate(config_data)


def save_config(name: str, model: BaseModel) -> None:
    """
    Save a configuration to a JSON file.

    Parameters
    ----------
    name : str
        The name of the configuration to save.
    model : Config
        The Pydantic model containing the configuration to save.

    Raises
    ------
    NotProvidedError
        If no name is provided for the configuration to save.
    ConfigExistsError
        If a configuration with the provided name already exists.

    """
    if not name:
        raise NotProvidedError(entity=Entity.CONFIG_NAME)

    CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    name = name.strip()
    config_path = CONFIG_DIR / f"{name}.json"
    if config_path.is_file():
        raise ConfigExistsError(name=name)
    json_dump(model.model_dump(), config_path)


def delete_configs(names: Sequence[str]) -> None:
    """
    Delete the configurations with the provided names.

    Parameters
    ----------
    names : Sequence[str]
        Names of the configurations to delete.

    Raises
    ------
    NotProvidedError
        If no names of items are provided.
    ConfigNotFoundError
        If a configuration with a provided name does not exist.

    """
    if not names:
        raise NotProvidedError(entity=Entity.CONFIG_NAMES, ui_msg=UIMessage.NO_CONFIG)
    config_file_paths: list[Path] = []
    for name in names:
        stripped_name = name.strip()
        config_file_path = Path(CONFIG_DIR) / f"{stripped_name}.json"
        if not config_file_path.is_file():
            raise ConfigNotFoundError(name=stripped_name)
        config_file_paths.append(config_file_path)
    for config_file_path in config_file_paths:
        config_file_path.unlink()


def delete_all_configs() -> None:
    """Delete all configuration files."""
    delete_directory(CONFIG_DIR)



"""Extra type definitions for the CLI of the Ultimate RVC project."""

from __future__ import annotations

from enum import StrEnum


class PanelName(StrEnum):
    """
    Valid panel names for audio generation commands in the CLI of
    the Ultimate RVC project.
    """

    MAIN_OPTIONS = "Main Options"
    RVC_MAIN_OPTIONS = "RVC Main Options"
    EDGE_TTS_OPTIONS = "Edge TTS Options"
    VOICE_SYNTHESIS_OPTIONS = "Voice Synthesis Options"
    RVC_SYNTHESIS_OPTIONS = "RVC Synthesis Options"
    RVC_VOICE_SYNTHESIS_OPTIONS = "RVC Voice Synthesis Options"
    VOICE_ENRICHMENT_OPTIONS = "Voice Enrichment Options"
    VOCAL_ENRICHMENT_OPTIONS = "Vocal Enrichment Options"
    RVC_ENRICHMENT_OPTIONS = "RVC Enrichment Options"
    SPEAKER_EMBEDDINGS_OPTIONS = "Speaker Embeddings Options"
    RVC_EMBEDDINGS_OPTIONS = "RVC Embeddings Options"
    VOCAL_POST_PROCESSING_OPTIONS = "Vocal Post-processing Options"
    AUDIO_MIXING_OPTIONS = "Audio Mixing Options"
    NETWORK_OPTIONS = "Network Options"
    EXTRACTION_OPTIONS = "Extraction Options"
    TRAINING_OPTIONS = "Training Options"
    ALGORITHMIC_OPTIONS = "Algorithmic Options"
    DATA_STORAGE_OPTIONS = "Data Storage Options"
    DEVICE_AND_MEMORY_OPTIONS = "Device and Memory Options"



import os

import torch


def change_info(path, info, name):
    try:
        ckpt = torch.load(path, map_location="cpu", weights_only=False)
        ckpt["info"] = info

        if not name:
            name = os.path.splitext(os.path.basename(path))[0]

        target_dir = os.path.join("logs", name)
        os.makedirs(target_dir, exist_ok=True)

        torch.save(ckpt, os.path.join(target_dir, f"{name}.pth"))

        return "Success."

    except Exception as error:
        print(f"An error occurred while changing the info: {error}")
        return f"Error: {error}"



from typing import IO, List, Optional, Tuple, Union

import json
import os
import re
import shutil
import sys
import tempfile
import textwrap
import time
import warnings
from urllib.parse import parse_qs, unquote, urlparse

import requests

from tqdm import tqdm

CHUNK_SIZE = 512 * 1024
HOME = os.path.expanduser("~")


def indent(text: str, prefix: str):
    """Indent each non-empty line of text with the given prefix."""
    return "".join(
        (prefix + line if line.strip() else line) for line in text.splitlines(True)
    )


class FileURLRetrievalError(Exception):
    """Custom exception for issues retrieving file URLs."""


def _extract_download_url_from_confirmation(contents: str, url_origin: str):
    """Extract the download URL from a Google Drive confirmation page."""
    patterns = [
        r'href="(\/uc\?export=download[^"]+)',
        r'href="/open\?id=([^"]+)"',
        r'"downloadUrl":"([^"]+)',
    ]
    for pattern in patterns:
        match = re.search(pattern, contents)
        if match:
            url = match.group(1)
            if pattern == r'href="/open\?id=([^"]+)"':
                uuid_match = re.search(
                    r'<input\s+type="hidden"\s+name="uuid"\s+value="([^"]+)"',
                    contents,
                )
                if uuid_match:
                    uuid = uuid_match.group(1)
                    return (
                        "https://drive.usercontent.google.com/download?id="
                        + url
                        + "&confirm=t&uuid="
                        + uuid
                    )
                raise FileURLRetrievalError(
                    f"Could not find UUID for download from {url_origin}",
                )
            if pattern == r'"downloadUrl":"([^"]+)':
                return url.replace("\\u003d", "=").replace("\\u0026", "&")
            return "https://docs.google.com" + url.replace("&", "&")

    error_match = re.search(r'<p class="uc-error-subcaption">(.*)</p>', contents)
    if error_match:
        error = error_match.group(1)
        raise FileURLRetrievalError(error)

    raise FileURLRetrievalError(
        "Cannot retrieve the public link of the file. "
        "You may need to change the permission to "
        "'Anyone with the link', or have had many accesses.",
    )


def _create_session(
    proxy: str | None = None,
    use_cookies: bool = True,
    return_cookies_file: bool = False,
):
    """Create a requests session with optional proxy and cookie handling."""
    sess = requests.session()
    sess.headers.update(
        {"User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6)"},
    )

    if proxy:
        sess.proxies = {"http": proxy, "https": proxy}

    cookies_file = os.path.join(HOME, ".cache/gdown/cookies.json")
    if os.path.exists(cookies_file) and use_cookies:
        try:
            with open(cookies_file) as f:
                cookies = json.load(f)
            for k, v in cookies:
                sess.cookies[k] = v
        except json.JSONDecodeError:
            warnings.warn("Corrupted Cookies file")

    return (sess, cookies_file) if return_cookies_file else sess


def download(
    output: str | None = None,
    quiet: bool = False,
    proxy: str | None = None,
    speed: float | None = None,
    use_cookies: bool = True,
    verify: bool | str = True,
    id: str | None = None,
    fuzzy: bool = True,
    resume: bool = False,
    format: str | None = None,
    url: str | None = None,
):
    """
    Download a file from a URL, supporting Google Drive links.

    Args:
        output: Output filepath. Default is basename of URL.
        quiet: Suppress terminal output.
        proxy: HTTP/HTTPS proxy.
        speed: Download speed limit (bytes per second).
        use_cookies: Flag to use cookies.
        verify: Verify TLS certificates.
        id: Google Drive's file ID.
        fuzzy: Fuzzy Google Drive ID extraction.
        resume: Resume download from a tmp file.
        format: Format for Google Docs/Sheets/Slides.
        url: URL to download from.

    Returns:
        Output filename, or None on error.

    """
    if not (id is None) ^ (url is None):
        raise ValueError("Either url or id has to be specified")

    if id is not None:
        url = f"https://drive.google.com/uc?id={id}"

    url_origin = url
    sess, cookies_file = _create_session(
        proxy=proxy,
        use_cookies=use_cookies,
        return_cookies_file=True,
    )

    while True:
        res = sess.get(url, stream=True, verify=verify)
        res.raise_for_status()

        if url == url_origin and res.status_code == 500:
            url = f"https://drive.google.com/open?id={id}"
            continue

        if res.headers.get("Content-Type", "").startswith("text/html"):
            title_match = re.search(r"<title>(.+)</title>", res.text)
            if title_match:
                title = title_match.group(1)
                if title.endswith(" - Google Docs"):
                    url = (
                        f"https://docs.google.com/document/d/{id}/export?format={'docx' if format is None else format}"
                    )
                    continue
                if title.endswith(" - Google Sheets"):
                    url = (
                        f"https://docs.google.com/spreadsheets/d/{id}/export?format={'xlsx' if format is None else format}"
                    )
                    continue
                if title.endswith(" - Google Slides"):
                    url = (
                        f"https://docs.google.com/presentation/d/{id}/export?format={'pptx' if format is None else format}"
                    )
                    continue
            if (
                "Content-Disposition" in res.headers
                and res.headers["Content-Disposition"].endswith("pptx")
                and format not in (None, "pptx")
            ):
                url = (
                    f"https://docs.google.com/presentation/d/{id}/export?format={'pptx' if format is None else format}"
                )
                continue

        if use_cookies:
            os.makedirs(os.path.dirname(cookies_file), exist_ok=True)
            cookies = [
                (k, v)
                for k, v in sess.cookies.items()
                if not k.startswith("download_warning_")
            ]
            with open(cookies_file, "w") as f:
                json.dump(cookies, f, indent=2)

        if "Content-Disposition" in res.headers:
            break

        parsed_url = urlparse(url)
        is_gdrive = parsed_url.hostname in ("drive.google.com", "docs.google.com")
        is_download_link = parsed_url.path.endswith("/uc")

        if not (is_gdrive and is_download_link and fuzzy):
            break

        try:
            url = _extract_download_url_from_confirmation(res.text, url_origin)
        except FileURLRetrievalError as e:
            raise FileURLRetrievalError(e)

    content_disposition = res.headers.get("Content-Disposition", "")
    filename_match = re.search(
        r"filename\*=UTF-8''(.*)",
        content_disposition,
    ) or re.search(r'filename=["\']?(.*?)["\']?$', content_disposition)
    filename_from_url = (
        unquote(filename_match.group(1)) if filename_match else os.path.basename(url)
    )
    download_path = output or filename_from_url

    if isinstance(download_path, str) and download_path.endswith(os.path.sep):
        os.makedirs(download_path, exist_ok=True)
        download_path = os.path.join(download_path, filename_from_url)

    temp_dir = os.path.dirname(download_path) or "."
    prefix = os.path.basename(download_path)

    if isinstance(download_path, str):
        existing_tmp_files = [
            os.path.join(temp_dir, file)
            for file in os.listdir(temp_dir)
            if file.startswith(prefix)
        ]
        if resume and existing_tmp_files:
            if len(existing_tmp_files) > 1:
                print(
                    "There are multiple temporary files to resume:",
                    file=sys.stderr,
                )
                for file in existing_tmp_files:
                    print(f"\t{file}", file=sys.stderr)
                print(
                    "Please remove them except one to resume downloading.",
                    file=sys.stderr,
                )
                return None
            temp_file_path = existing_tmp_files[0]
        else:
            resume = False
            temp_file_path = tempfile.mktemp(
                suffix=tempfile.template,
                prefix=prefix,
                dir=temp_dir,
            )

        try:
            file_obj: IO = open(temp_file_path, "ab")
        except Exception as e:
            print(
                f"Could not open the temporary file {temp_file_path}: {e}",
                file=sys.stderr,
            )
            return None
    else:
        temp_file_path = None
        file_obj = download_path

    if temp_file_path is not None and file_obj.tell() != 0:
        headers = {"Range": f"bytes={file_obj.tell()}-"}
        res = sess.get(url, headers=headers, stream=True, verify=verify)
        res.raise_for_status()

    try:
        total = int(res.headers.get("Content-Length", 0))
        if total > 0:
            if not quiet:
                pbar = tqdm(
                    total=total,
                    unit="B",
                    unit_scale=True,
                    desc=filename_from_url,
                )
        elif not quiet:
            pbar = tqdm(unit="B", unit_scale=True, desc=filename_from_url)

        t_start = time.time()
        for chunk in res.iter_content(chunk_size=CHUNK_SIZE):
            file_obj.write(chunk)
            if not quiet:
                pbar.update(len(chunk))
            if speed is not None:
                elapsed_time_expected = 1.0 * pbar.n / speed
                elapsed_time = time.time() - t_start
                if elapsed_time < elapsed_time_expected:
                    time.sleep(elapsed_time_expected - elapsed_time)
        if not quiet:
            pbar.close()

        if temp_file_path:
            file_obj.close()
            shutil.move(temp_file_path, download_path)
    finally:
        sess.close()

    return download_path



import numpy as np

import librosa


def process_audio(audio, sr=16000, silence_thresh=-60, min_silence_len=250):
    """
    Splits an audio signal into segments using a fixed frame size and hop size.

    Parameters
    ----------
    - audio (np.ndarray): The audio signal to split.
    - sr (int): The sample rate of the input audio (default is 16000).
    - silence_thresh (int): Silence threshold (default =-60dB)
    - min_silence_len (int): Minimum silence duration (default 250ms).

    Returns
    -------
    - list of np.ndarray: A list of audio segments.
    - np.ndarray: The intervals where the audio was split.

    """
    frame_length = int(min_silence_len / 1000 * sr)
    hop_length = frame_length // 2
    intervals = librosa.effects.split(
        audio,
        top_db=-silence_thresh,
        frame_length=frame_length,
        hop_length=hop_length,
    )
    audio_segments = [audio[start:end] for start, end in intervals]

    return audio_segments, intervals


def merge_audio(audio_segments_org, audio_segments_new, intervals, sr_orig, sr_new):
    """
    Merges audio segments back into a single audio signal, filling gaps with silence.
    Assumes audio segments are already at sr_new.

    Parameters
    ----------
    - audio_segments_org (list of np.ndarray): The non-silent audio segments (at sr_orig).
    - audio_segments_new (list of np.ndarray): The non-silent audio segments (at sr_new).
    - intervals (np.ndarray): The intervals used for splitting the original audio.
    - sr_orig (int): The sample rate of the original audio
    - sr_new (int): The sample rate of the model
    Returns:
    - np.ndarray: The merged audio signal with silent gaps restored.

    """
    merged_audio = np.array([], dtype=audio_segments_new[0].dtype)
    sr_ratio = sr_new / sr_orig

    for i, (start, end) in enumerate(intervals):

        start_new = int(start * sr_ratio)
        end_new = int(end * sr_ratio)

        original_duration = len(audio_segments_org[i]) / sr_orig
        new_duration = len(audio_segments_new[i]) / sr_new
        duration_diff = new_duration - original_duration

        silence_samples = int(abs(duration_diff) * sr_new)
        silence_compensation = np.zeros(
            silence_samples,
            dtype=audio_segments_new[0].dtype,
        )

        if i == 0 and start_new > 0:
            initial_silence = np.zeros(start_new, dtype=audio_segments_new[0].dtype)
            merged_audio = np.concatenate((merged_audio, initial_silence))

        if duration_diff > 0:
            merged_audio = np.concatenate((merged_audio, silence_compensation))

        merged_audio = np.concatenate((merged_audio, audio_segments_new[i]))

        if duration_diff < 0:
            merged_audio = np.concatenate((merged_audio, silence_compensation))

        if i < len(intervals) - 1:
            next_start_new = int(intervals[i + 1][0] * sr_ratio)
            silence_duration = next_start_new - end_new
            if silence_duration > 0:
                silence = np.zeros(silence_duration, dtype=audio_segments_new[0].dtype)
                merged_audio = np.concatenate((merged_audio, silence))

    return merged_audio
